# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***
from __future__ import annotations

import functools
from pathlib import Path
from typing import overload, List  # noqa
from nptyping import NDArray, Float  # noqa

from jpype import JException
from jpype.types import JBoolean, JInt, JFloat

from .pmath import _get_pvector_wrapper  # noqa


def _return_list_py5shapes(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        return [Py5Shape(s) for s in f(self_, *args)]
    return decorated


def _return_py5shape(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        return Py5Shape(f(self_, *args))
    return decorated


def _py5shape_type_fixer(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        args = list(args)

        def fix_type(arg):
            if isinstance(arg, bool):
                return JBoolean(arg)
            elif isinstance(arg, int):
                return JInt(arg)
            elif isinstance(arg, float):
                return JFloat(arg)
            else:
                return arg
        args = [fix_type(a) for a in args]
        return f(self_, *tuple(args))
    return decorated


def _load_py5shape(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        try:
            return Py5Shape(f(self_, *args))
        except JException as e:
            msg = e.message()
            if msg == 'None':
                msg = 'shape file cannot be found'
        raise RuntimeError('cannot load shape ' +
                           str(args[0]) + '. error message: ' + msg)
    return decorated


class Py5Shape:
    """Datatype for storing shapes.

    Underlying Java class: PShape.PShape

    Notes
    -----

    Datatype for storing shapes. Before a shape is used, it must be loaded with the
    ``load_shape()`` or created with the ``create_shape()``. The ``shape()``
    function is used to draw the shape to the display window. Processing can
    currently load and display SVG (Scalable Vector Graphics) and OBJ shapes. OBJ
    files can only be opened using the ``P3D`` renderer. The ``load_shape()``
    function supports SVG files created with Inkscape and Adobe Illustrator. It is
    not a full SVG implementation, but offers some straightforward support for
    handling vector data.

    The ``Py5Shape`` object contains a group of methods that can operate on the
    shape data. Some of the methods are listed below, but the full list used for
    creating and modifying shapes is available here in the Processing Javadoc.

    To create a new shape, use the ``create_shape()`` function. Do not use the
    syntax ``new Py5Shape()``.
"""

    def __init__(self, pshape):
        self._instance = pshape

    def _get_depth(self) -> float:
        """new template no description.

        Underlying Java field: PShape.depth

        Notes
        -----

        new template no description.
"""
        return self._instance.depth
    depth: float = property(fget=_get_depth)

    def _get_height(self) -> float:
        """The height of the PShape document.

        Underlying Java field: PShape.height

        Notes
        -----

        The height of the PShape document.
"""
        return self._instance.height
    height: float = property(fget=_get_height)

    def _get_width(self) -> float:
        """The width of the PShape document.

        Underlying Java field: PShape.width

        Notes
        -----

        The width of the PShape document.
"""
        return self._instance.width
    width: float = property(fget=_get_width)

    @overload
    def add_child(self, who: Py5Shape, /) -> None:
        """Adds a child PShape to a parent PShape that is defined as a GROUP.

        Underlying Java method: PShape.addChild

        Methods
        -------

        You can use any of the following signatures:

         * add_child(who: Py5Shape) -> None
         * add_child(who: Py5Shape, idx: int) -> None

        Parameters
        ----------

        idx: int
            the layer position in which to insert the new child

        who: Py5Shape
            any variable of type PShape

        Notes
        -----

        Adds a child PShape to a parent PShape that is defined as a GROUP. In the
        example, the three shapes ``path``, ``rectangle``, and ``circle`` are added to a
        parent PShape variable named ``house`` that is a GROUP.
"""
        pass

    @overload
    def add_child(self, who: Py5Shape, idx: int, /) -> None:
        """Adds a child PShape to a parent PShape that is defined as a GROUP.

        Underlying Java method: PShape.addChild

        Methods
        -------

        You can use any of the following signatures:

         * add_child(who: Py5Shape) -> None
         * add_child(who: Py5Shape, idx: int) -> None

        Parameters
        ----------

        idx: int
            the layer position in which to insert the new child

        who: Py5Shape
            any variable of type PShape

        Notes
        -----

        Adds a child PShape to a parent PShape that is defined as a GROUP. In the
        example, the three shapes ``path``, ``rectangle``, and ``circle`` are added to a
        parent PShape variable named ``house`` that is a GROUP.
"""
        pass

    def add_child(self, *args):
        """Adds a child PShape to a parent PShape that is defined as a GROUP.

        Underlying Java method: PShape.addChild

        Methods
        -------

        You can use any of the following signatures:

         * add_child(who: Py5Shape) -> None
         * add_child(who: Py5Shape, idx: int) -> None

        Parameters
        ----------

        idx: int
            the layer position in which to insert the new child

        who: Py5Shape
            any variable of type PShape

        Notes
        -----

        Adds a child PShape to a parent PShape that is defined as a GROUP. In the
        example, the three shapes ``path``, ``rectangle``, and ``circle`` are added to a
        parent PShape variable named ``house`` that is a GROUP.
"""
        return self._instance.addChild(*args)

    def add_name(self, nom: str, shape: Py5Shape, /) -> None:
        """new template no description.

        Underlying Java method: PShape.addName

        Parameters
        ----------

        nom: str
            missing variable description

        shape: Py5Shape
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.addName(nom, shape)

    @overload
    def apply_matrix(self, n00: float, n01: float, n02: float,
                     n10: float, n11: float, n12: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.applyMatrix

        Methods
        -------

        You can use any of the following signatures:

         * apply_matrix(n00: float, n01: float, n02: float, n03: float, n10: float, n11: float, n12: float, n13: float, n20: float, n21: float, n22: float, n23: float, n30: float, n31: float, n32: float, n33: float) -> None
         * apply_matrix(n00: float, n01: float, n02: float, n10: float, n11: float, n12: float) -> None
         * apply_matrix(source: NDArray[(2, 3), Float]) -> None
         * apply_matrix(source: NDArray[(4, 4), Float]) -> None

        Parameters
        ----------

        n00: float
            missing variable description

        n01: float
            missing variable description

        n02: float
            missing variable description

        n03: float
            missing variable description

        n10: float
            missing variable description

        n11: float
            missing variable description

        n12: float
            missing variable description

        n13: float
            missing variable description

        n20: float
            missing variable description

        n21: float
            missing variable description

        n22: float
            missing variable description

        n23: float
            missing variable description

        n30: float
            missing variable description

        n31: float
            missing variable description

        n32: float
            missing variable description

        n33: float
            missing variable description

        source: NDArray[(2, 3), Float]
            missing variable description

        source: NDArray[(4, 4), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def apply_matrix(
            self,
            n00: float,
            n01: float,
            n02: float,
            n03: float,
            n10: float,
            n11: float,
            n12: float,
            n13: float,
            n20: float,
            n21: float,
            n22: float,
            n23: float,
            n30: float,
            n31: float,
            n32: float,
            n33: float,
            /) -> None:
        """new template no description.

        Underlying Java method: PShape.applyMatrix

        Methods
        -------

        You can use any of the following signatures:

         * apply_matrix(n00: float, n01: float, n02: float, n03: float, n10: float, n11: float, n12: float, n13: float, n20: float, n21: float, n22: float, n23: float, n30: float, n31: float, n32: float, n33: float) -> None
         * apply_matrix(n00: float, n01: float, n02: float, n10: float, n11: float, n12: float) -> None
         * apply_matrix(source: NDArray[(2, 3), Float]) -> None
         * apply_matrix(source: NDArray[(4, 4), Float]) -> None

        Parameters
        ----------

        n00: float
            missing variable description

        n01: float
            missing variable description

        n02: float
            missing variable description

        n03: float
            missing variable description

        n10: float
            missing variable description

        n11: float
            missing variable description

        n12: float
            missing variable description

        n13: float
            missing variable description

        n20: float
            missing variable description

        n21: float
            missing variable description

        n22: float
            missing variable description

        n23: float
            missing variable description

        n30: float
            missing variable description

        n31: float
            missing variable description

        n32: float
            missing variable description

        n33: float
            missing variable description

        source: NDArray[(2, 3), Float]
            missing variable description

        source: NDArray[(4, 4), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def apply_matrix(self, source: NDArray[(2, 3), Float], /) -> None:
        """new template no description.

        Underlying Java method: PShape.applyMatrix

        Methods
        -------

        You can use any of the following signatures:

         * apply_matrix(n00: float, n01: float, n02: float, n03: float, n10: float, n11: float, n12: float, n13: float, n20: float, n21: float, n22: float, n23: float, n30: float, n31: float, n32: float, n33: float) -> None
         * apply_matrix(n00: float, n01: float, n02: float, n10: float, n11: float, n12: float) -> None
         * apply_matrix(source: NDArray[(2, 3), Float]) -> None
         * apply_matrix(source: NDArray[(4, 4), Float]) -> None

        Parameters
        ----------

        n00: float
            missing variable description

        n01: float
            missing variable description

        n02: float
            missing variable description

        n03: float
            missing variable description

        n10: float
            missing variable description

        n11: float
            missing variable description

        n12: float
            missing variable description

        n13: float
            missing variable description

        n20: float
            missing variable description

        n21: float
            missing variable description

        n22: float
            missing variable description

        n23: float
            missing variable description

        n30: float
            missing variable description

        n31: float
            missing variable description

        n32: float
            missing variable description

        n33: float
            missing variable description

        source: NDArray[(2, 3), Float]
            missing variable description

        source: NDArray[(4, 4), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def apply_matrix(self, source: NDArray[(4, 4), Float], /) -> None:
        """new template no description.

        Underlying Java method: PShape.applyMatrix

        Methods
        -------

        You can use any of the following signatures:

         * apply_matrix(n00: float, n01: float, n02: float, n03: float, n10: float, n11: float, n12: float, n13: float, n20: float, n21: float, n22: float, n23: float, n30: float, n31: float, n32: float, n33: float) -> None
         * apply_matrix(n00: float, n01: float, n02: float, n10: float, n11: float, n12: float) -> None
         * apply_matrix(source: NDArray[(2, 3), Float]) -> None
         * apply_matrix(source: NDArray[(4, 4), Float]) -> None

        Parameters
        ----------

        n00: float
            missing variable description

        n01: float
            missing variable description

        n02: float
            missing variable description

        n03: float
            missing variable description

        n10: float
            missing variable description

        n11: float
            missing variable description

        n12: float
            missing variable description

        n13: float
            missing variable description

        n20: float
            missing variable description

        n21: float
            missing variable description

        n22: float
            missing variable description

        n23: float
            missing variable description

        n30: float
            missing variable description

        n31: float
            missing variable description

        n32: float
            missing variable description

        n33: float
            missing variable description

        source: NDArray[(2, 3), Float]
            missing variable description

        source: NDArray[(4, 4), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def apply_matrix(self, *args):
        """new template no description.

        Underlying Java method: PShape.applyMatrix

        Methods
        -------

        You can use any of the following signatures:

         * apply_matrix(n00: float, n01: float, n02: float, n03: float, n10: float, n11: float, n12: float, n13: float, n20: float, n21: float, n22: float, n23: float, n30: float, n31: float, n32: float, n33: float) -> None
         * apply_matrix(n00: float, n01: float, n02: float, n10: float, n11: float, n12: float) -> None
         * apply_matrix(source: NDArray[(2, 3), Float]) -> None
         * apply_matrix(source: NDArray[(4, 4), Float]) -> None

        Parameters
        ----------

        n00: float
            missing variable description

        n01: float
            missing variable description

        n02: float
            missing variable description

        n03: float
            missing variable description

        n10: float
            missing variable description

        n11: float
            missing variable description

        n12: float
            missing variable description

        n13: float
            missing variable description

        n20: float
            missing variable description

        n21: float
            missing variable description

        n22: float
            missing variable description

        n23: float
            missing variable description

        n30: float
            missing variable description

        n31: float
            missing variable description

        n32: float
            missing variable description

        n33: float
            missing variable description

        source: NDArray[(2, 3), Float]
            missing variable description

        source: NDArray[(4, 4), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.applyMatrix(*args)

    @overload
    def attrib(self, name: str, /, *values: bool) -> None:
        """new template no description.

        Underlying Java method: PShape.attrib

        Methods
        -------

        You can use any of the following signatures:

         * attrib(name: str, values: bool) -> None
         * attrib(name: str, values: float) -> None
         * attrib(name: str, values: int) -> None

        Parameters
        ----------

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def attrib(self, name: str, /, *values: float) -> None:
        """new template no description.

        Underlying Java method: PShape.attrib

        Methods
        -------

        You can use any of the following signatures:

         * attrib(name: str, values: bool) -> None
         * attrib(name: str, values: float) -> None
         * attrib(name: str, values: int) -> None

        Parameters
        ----------

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def attrib(self, name: str, /, *values: int) -> None:
        """new template no description.

        Underlying Java method: PShape.attrib

        Methods
        -------

        You can use any of the following signatures:

         * attrib(name: str, values: bool) -> None
         * attrib(name: str, values: float) -> None
         * attrib(name: str, values: int) -> None

        Parameters
        ----------

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @_py5shape_type_fixer
    def attrib(self, *args):
        """new template no description.

        Underlying Java method: PShape.attrib

        Methods
        -------

        You can use any of the following signatures:

         * attrib(name: str, values: bool) -> None
         * attrib(name: str, values: float) -> None
         * attrib(name: str, values: int) -> None

        Parameters
        ----------

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.attrib(*args)

    def attrib_color(self, name: str, color: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.attribColor

        Parameters
        ----------

        color: int
            missing variable description

        name: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.attribColor(name, color)

    def attrib_normal(self, name: str, nx: float,
                      ny: float, nz: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.attribNormal

        Parameters
        ----------

        name: str
            missing variable description

        nx: float
            missing variable description

        ny: float
            missing variable description

        nz: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.attribNormal(name, nx, ny, nz)

    def attrib_position(self, name: str, x: float,
                        y: float, z: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.attribPosition

        Parameters
        ----------

        name: str
            missing variable description

        x: float
            missing variable description

        y: float
            missing variable description

        z: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.attribPosition(name, x, y, z)

    @overload
    def begin_shape(self) -> None:
        """This method is used to start a custom shape created with the ``create_shape()``
        function.

        Underlying Java method: PShape.beginShape

        Methods
        -------

        You can use any of the following signatures:

         * begin_shape() -> None
         * begin_shape(kind: int) -> None

        Parameters
        ----------

        kind: int
            missing variable description

        Notes
        -----

        This method is used to start a custom shape created with the ``create_shape()``
        function. It's always and only used with ``create_shape()``.
"""
        pass

    @overload
    def begin_shape(self, kind: int, /) -> None:
        """This method is used to start a custom shape created with the ``create_shape()``
        function.

        Underlying Java method: PShape.beginShape

        Methods
        -------

        You can use any of the following signatures:

         * begin_shape() -> None
         * begin_shape(kind: int) -> None

        Parameters
        ----------

        kind: int
            missing variable description

        Notes
        -----

        This method is used to start a custom shape created with the ``create_shape()``
        function. It's always and only used with ``create_shape()``.
"""
        pass

    def begin_shape(self, *args):
        """This method is used to start a custom shape created with the ``create_shape()``
        function.

        Underlying Java method: PShape.beginShape

        Methods
        -------

        You can use any of the following signatures:

         * begin_shape() -> None
         * begin_shape(kind: int) -> None

        Parameters
        ----------

        kind: int
            missing variable description

        Notes
        -----

        This method is used to start a custom shape created with the ``create_shape()``
        function. It's always and only used with ``create_shape()``.
"""
        return self._instance.beginShape(*args)

    def bezier_detail(self, detail: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.bezierDetail

        Parameters
        ----------

        detail: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.bezierDetail(detail)

    def contains(self, x: float, y: float, /) -> bool:
        """new template no description.

        Underlying Java method: PShape.contains

        Parameters
        ----------

        x: float
            missing variable description

        y: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.contains(x, y)

    def disable_style(self) -> None:
        """Disables the shape's style data and uses Processing's current styles.

        Underlying Java method: PShape.disableStyle

        Notes
        -----

        Disables the shape's style data and uses Processing's current styles. Styles
        include attributes such as colors, stroke weight, and stroke joints.
"""
        return self._instance.disableStyle()

    def enable_style(self) -> None:
        """Enables the shape's style data and ignores Processing's current styles.

        Underlying Java method: PShape.enableStyle

        Notes
        -----

        Enables the shape's style data and ignores Processing's current styles. Styles
        include attributes such as colors, stroke weight, and stroke joints.
"""
        return self._instance.enableStyle()

    @overload
    def end_shape(self) -> None:
        """This method is used to complete a custom shape created with the
        ``create_shape()`` function.

        Underlying Java method: PShape.endShape

        Methods
        -------

        You can use any of the following signatures:

         * end_shape() -> None
         * end_shape(mode: int) -> None

        Parameters
        ----------

        mode: int
            missing variable description

        Notes
        -----

        This method is used to complete a custom shape created with the
        ``create_shape()`` function. It's always and only used with ``create_shape()``.
"""
        pass

    @overload
    def end_shape(self, mode: int, /) -> None:
        """This method is used to complete a custom shape created with the
        ``create_shape()`` function.

        Underlying Java method: PShape.endShape

        Methods
        -------

        You can use any of the following signatures:

         * end_shape() -> None
         * end_shape(mode: int) -> None

        Parameters
        ----------

        mode: int
            missing variable description

        Notes
        -----

        This method is used to complete a custom shape created with the
        ``create_shape()`` function. It's always and only used with ``create_shape()``.
"""
        pass

    def end_shape(self, *args):
        """This method is used to complete a custom shape created with the
        ``create_shape()`` function.

        Underlying Java method: PShape.endShape

        Methods
        -------

        You can use any of the following signatures:

         * end_shape() -> None
         * end_shape(mode: int) -> None

        Parameters
        ----------

        mode: int
            missing variable description

        Notes
        -----

        This method is used to complete a custom shape created with the
        ``create_shape()`` function. It's always and only used with ``create_shape()``.
"""
        return self._instance.endShape(*args)

    @_return_py5shape
    def find_child(self, target: str, /) -> Py5Shape:
        """new template no description.

        Underlying Java method: PShape.findChild

        Parameters
        ----------

        target: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.findChild(target)

    def get_ambient(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getAmbient

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getAmbient(index)

    @overload
    def get_child(self, index: int, /) -> Py5Shape:
        """Extracts a child shape from a parent shape.

        Underlying Java method: PShape.getChild

        Methods
        -------

        You can use any of the following signatures:

         * get_child(index: int) -> Py5Shape
         * get_child(target: str) -> Py5Shape

        Parameters
        ----------

        index: int
            the layer position of the shape to get

        target: str
            the name of the shape to get

        Notes
        -----

        Extracts a child shape from a parent shape. Specify the name of the shape with
        the ``target`` parameter. The shape is returned as a ``Py5Shape`` object, or
        ``None`` is returned if there is an error.
"""
        pass

    @overload
    def get_child(self, target: str, /) -> Py5Shape:
        """Extracts a child shape from a parent shape.

        Underlying Java method: PShape.getChild

        Methods
        -------

        You can use any of the following signatures:

         * get_child(index: int) -> Py5Shape
         * get_child(target: str) -> Py5Shape

        Parameters
        ----------

        index: int
            the layer position of the shape to get

        target: str
            the name of the shape to get

        Notes
        -----

        Extracts a child shape from a parent shape. Specify the name of the shape with
        the ``target`` parameter. The shape is returned as a ``Py5Shape`` object, or
        ``None`` is returned if there is an error.
"""
        pass

    @_return_py5shape
    def get_child(self, *args):
        """Extracts a child shape from a parent shape.

        Underlying Java method: PShape.getChild

        Methods
        -------

        You can use any of the following signatures:

         * get_child(index: int) -> Py5Shape
         * get_child(target: str) -> Py5Shape

        Parameters
        ----------

        index: int
            the layer position of the shape to get

        target: str
            the name of the shape to get

        Notes
        -----

        Extracts a child shape from a parent shape. Specify the name of the shape with
        the ``target`` parameter. The shape is returned as a ``Py5Shape`` object, or
        ``None`` is returned if there is an error.
"""
        return self._instance.getChild(*args)

    def get_child_count(self) -> int:
        """Returns the number of children within the PShape.

        Underlying Java method: PShape.getChildCount

        Notes
        -----

        Returns the number of children within the PShape.
"""
        return self._instance.getChildCount()

    def get_child_index(self, who: Py5Shape, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getChildIndex

        Parameters
        ----------

        who: Py5Shape
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getChildIndex(who)

    @_return_list_py5shapes
    def get_children(self) -> List[Py5Shape]:
        """new template no description.

        Underlying Java method: PShape.getChildren

        Notes
        -----

        new template no description.
"""
        return self._instance.getChildren()

    def get_depth(self) -> float:
        """new template no description.

        Underlying Java method: PShape.getDepth

        Notes
        -----

        new template no description.
"""
        return self._instance.getDepth()

    def get_emissive(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getEmissive

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getEmissive(index)

    def get_family(self) -> int:
        """new template no description.

        Underlying Java method: PShape.getFamily

        Notes
        -----

        new template no description.
"""
        return self._instance.getFamily()

    def get_fill(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getFill

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getFill(index)

    def get_height(self) -> float:
        """new template no description.

        Underlying Java method: PShape.getHeight

        Notes
        -----

        new template no description.
"""
        return self._instance.getHeight()

    def get_kind(self) -> int:
        """new template no description.

        Underlying Java method: PShape.getKind

        Notes
        -----

        new template no description.
"""
        return self._instance.getKind()

    def get_name(self) -> str:
        """new template no description.

        Underlying Java method: PShape.getName

        Notes
        -----

        new template no description.
"""
        return self._instance.getName()

    @overload
    def get_normal(self, index: int, /) -> NDArray[(Any,), Float]:
        """new template no description.

        Underlying Java method: PShape.getNormal

        Methods
        -------

        You can use any of the following signatures:

         * get_normal(index: int) -> NDArray[(Any,), Float]
         * get_normal(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            missing variable description

        vec: NDArray[(Any,), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def get_normal(self, index: int, vec: NDArray[(
            Any,), Float], /) -> NDArray[(Any,), Float]:
        """new template no description.

        Underlying Java method: PShape.getNormal

        Methods
        -------

        You can use any of the following signatures:

         * get_normal(index: int) -> NDArray[(Any,), Float]
         * get_normal(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            missing variable description

        vec: NDArray[(Any,), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @_get_pvector_wrapper
    def get_normal(self, *args):
        """new template no description.

        Underlying Java method: PShape.getNormal

        Methods
        -------

        You can use any of the following signatures:

         * get_normal(index: int) -> NDArray[(Any,), Float]
         * get_normal(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            missing variable description

        vec: NDArray[(Any,), Float]
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getNormal(*args)

    def get_normal_x(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getNormalX

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getNormalX(index)

    def get_normal_y(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getNormalY

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getNormalY(index)

    def get_normal_z(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getNormalZ

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getNormalZ(index)

    @_return_py5shape
    def get_parent(self) -> Py5Shape:
        """new template no description.

        Underlying Java method: PShape.getParent

        Notes
        -----

        new template no description.
"""
        return self._instance.getParent()

    def get_shininess(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getShininess

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getShininess(index)

    def get_specular(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getSpecular

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getSpecular(index)

    def get_stroke(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getStroke

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getStroke(index)

    def get_stroke_weight(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getStrokeWeight

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getStrokeWeight(index)

    @_return_py5shape
    def get_tessellation(self) -> Py5Shape:
        """new template no description.

        Underlying Java method: PShape.getTessellation

        Notes
        -----

        new template no description.
"""
        return self._instance.getTessellation()

    def get_texture_u(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getTextureU

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getTextureU(index)

    def get_texture_v(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getTextureV

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getTextureV(index)

    def get_tint(self, index: int, /) -> int:
        """new template no description.

        Underlying Java method: PShape.getTint

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getTint(index)

    @overload
    def get_vertex(self, index: int, /) -> NDArray[(Any,), Float]:
        """The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.getVertex

        Methods
        -------

        You can use any of the following signatures:

         * get_vertex(index: int) -> NDArray[(Any,), Float]
         * get_vertex(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            PVector to assign the data to

        Notes
        -----

        The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter. This method
        works when shapes are created as shown in the example above, but won't work
        properly when a shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20,
        80, 80)``.
"""
        pass

    @overload
    def get_vertex(self, index: int, vec: NDArray[(
            Any,), Float], /) -> NDArray[(Any,), Float]:
        """The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.getVertex

        Methods
        -------

        You can use any of the following signatures:

         * get_vertex(index: int) -> NDArray[(Any,), Float]
         * get_vertex(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            PVector to assign the data to

        Notes
        -----

        The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter. This method
        works when shapes are created as shown in the example above, but won't work
        properly when a shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20,
        80, 80)``.
"""
        pass

    @_get_pvector_wrapper
    def get_vertex(self, *args):
        """The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.getVertex

        Methods
        -------

        You can use any of the following signatures:

         * get_vertex(index: int) -> NDArray[(Any,), Float]
         * get_vertex(index: int, vec: NDArray[(Any,), Float]) -> NDArray[(Any,), Float]

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            PVector to assign the data to

        Notes
        -----

        The ``get_vertex()`` method returns a PVector with the coordinates of the vertex
        point located at the position defined by the ``index`` parameter. This method
        works when shapes are created as shown in the example above, but won't work
        properly when a shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20,
        80, 80)``.
"""
        return self._instance.getVertex(*args)

    def get_vertex_count(self) -> int:
        """The ``get_vertex_count()`` method returns the number of vertices that make up a
        PShape.

        Underlying Java method: PShape.getVertexCount

        Notes
        -----

        The ``get_vertex_count()`` method returns the number of vertices that make up a
        PShape. In the above example, the value 4 is returned by the
        ``get_vertex_count()`` method because 4 vertices are defined in ``setup()``.
"""
        return self._instance.getVertexCount()

    def get_vertex_x(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getVertexX

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getVertexX(index)

    def get_vertex_y(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getVertexY

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getVertexY(index)

    def get_vertex_z(self, index: int, /) -> float:
        """new template no description.

        Underlying Java method: PShape.getVertexZ

        Parameters
        ----------

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getVertexZ(index)

    def get_width(self) -> float:
        """new template no description.

        Underlying Java method: PShape.getWidth

        Notes
        -----

        new template no description.
"""
        return self._instance.getWidth()

    def is2d(self) -> bool:
        """new template no description.

        Underlying Java method: PShape.is2D

        Notes
        -----

        new template no description.
"""
        return self._instance.is2D()

    def is3d(self) -> bool:
        """new template no description.

        Underlying Java method: PShape.is3D

        Notes
        -----

        new template no description.
"""
        return self._instance.is3D()

    def is_closed(self) -> bool:
        """new template no description.

        Underlying Java method: PShape.isClosed

        Notes
        -----

        new template no description.
"""
        return self._instance.isClosed()

    def is_visible(self) -> bool:
        """Returns a boolean value "true" if the image is set to be visible, "false" if
        not.

        Underlying Java method: PShape.isVisible

        Notes
        -----

        Returns a boolean value "true" if the image is set to be visible, "false" if
        not. This value can be modified with the ``set_visible()`` method.

        The default visibility of a shape is usually controlled by whatever program
        created the SVG file. For instance, this parameter is controlled by showing or
        hiding the shape in the layers palette in Adobe Illustrator.
"""
        return self._instance.isVisible()

    def remove_child(self, idx: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.removeChild

        Parameters
        ----------

        idx: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.removeChild(idx)

    def reset_matrix(self) -> None:
        """Replaces the current matrix of a shape with the identity matrix.

        Underlying Java method: PShape.resetMatrix

        Notes
        -----

        Replaces the current matrix of a shape with the identity matrix. The equivalent
        function in OpenGL is ``gl_load_identity()``.
"""
        return self._instance.resetMatrix()

    @overload
    def rotate(self, angle: float, /) -> None:
        """Rotates the shape the amount specified by the ``angle`` parameter.

        Underlying Java method: PShape.rotate

        Methods
        -------

        You can use any of the following signatures:

         * rotate(angle: float) -> None
         * rotate(angle: float, v0: float, v1: float, v2: float) -> None

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        v0: float
            missing variable description

        v1: float
            missing variable description

        v2: float
            missing variable description

        Notes
        -----

        Rotates the shape the amount specified by the ``angle`` parameter. Angles should
        be specified in radians (values from 0 to TWO_PI) or converted from degrees to
        radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Transformations apply
        to everything that happens after and subsequent calls to the method accumulates
        the effect. For example, calling ``rotate(HALF_PI)`` and then
        ``rotate(HALF_PI)`` is the same as ``rotate(PI)``. This transformation is
        applied directly to the shape, it's not refreshed each time ``draw()`` is run.
"""
        pass

    @overload
    def rotate(self, angle: float, v0: float, v1: float, v2: float, /) -> None:
        """Rotates the shape the amount specified by the ``angle`` parameter.

        Underlying Java method: PShape.rotate

        Methods
        -------

        You can use any of the following signatures:

         * rotate(angle: float) -> None
         * rotate(angle: float, v0: float, v1: float, v2: float) -> None

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        v0: float
            missing variable description

        v1: float
            missing variable description

        v2: float
            missing variable description

        Notes
        -----

        Rotates the shape the amount specified by the ``angle`` parameter. Angles should
        be specified in radians (values from 0 to TWO_PI) or converted from degrees to
        radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Transformations apply
        to everything that happens after and subsequent calls to the method accumulates
        the effect. For example, calling ``rotate(HALF_PI)`` and then
        ``rotate(HALF_PI)`` is the same as ``rotate(PI)``. This transformation is
        applied directly to the shape, it's not refreshed each time ``draw()`` is run.
"""
        pass

    def rotate(self, *args):
        """Rotates the shape the amount specified by the ``angle`` parameter.

        Underlying Java method: PShape.rotate

        Methods
        -------

        You can use any of the following signatures:

         * rotate(angle: float) -> None
         * rotate(angle: float, v0: float, v1: float, v2: float) -> None

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        v0: float
            missing variable description

        v1: float
            missing variable description

        v2: float
            missing variable description

        Notes
        -----

        Rotates the shape the amount specified by the ``angle`` parameter. Angles should
        be specified in radians (values from 0 to TWO_PI) or converted from degrees to
        radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Transformations apply
        to everything that happens after and subsequent calls to the method accumulates
        the effect. For example, calling ``rotate(HALF_PI)`` and then
        ``rotate(HALF_PI)`` is the same as ``rotate(PI)``. This transformation is
        applied directly to the shape, it's not refreshed each time ``draw()`` is run.
"""
        return self._instance.rotate(*args)

    def rotate_x(self, angle: float, /) -> None:
        """Rotates the shape around the x-axis the amount specified by the ``angle``
        parameter.

        Underlying Java method: PShape.rotateX

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        Notes
        -----

        Rotates the shape around the x-axis the amount specified by the ``angle``
        parameter. Angles should be specified in radians (values from 0 to TWO_PI) or
        converted from degrees to radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Subsequent calls to
        the method accumulates the effect. For example, calling ``rotate_x(HALF_PI)``
        and then ``rotate_x(HALF_PI)`` is the same as ``rotate_x(PI)``. This
        transformation is applied directly to the shape, it's not refreshed each time
        ``draw()`` is run.

        This method requires a 3D renderer. You need to use P3D as a third parameter for
        the ``size()`` function as shown in the example above.
"""
        return self._instance.rotateX(angle)

    def rotate_y(self, angle: float, /) -> None:
        """Rotates the shape around the y-axis the amount specified by the ``angle``
        parameter.

        Underlying Java method: PShape.rotateY

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        Notes
        -----

        Rotates the shape around the y-axis the amount specified by the ``angle``
        parameter. Angles should be specified in radians (values from 0 to TWO_PI) or
        converted from degrees to radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Subsequent calls to
        the method accumulates the effect. For example, calling ``rotate_y(HALF_PI)``
        and then ``rotate_y(HALF_PI)`` is the same as ``rotate_y(PI)``. This
        transformation is applied directly to the shape, it's not refreshed each time
        ``draw()`` is run.

        This method requires a 3D renderer. You need to use P3D as a third parameter for
        the ``size()`` function as shown in the example above.
"""
        return self._instance.rotateY(angle)

    def rotate_z(self, angle: float, /) -> None:
        """Rotates the shape around the z-axis the amount specified by the ``angle``
        parameter.

        Underlying Java method: PShape.rotateZ

        Parameters
        ----------

        angle: float
            angle of rotation specified in radians

        Notes
        -----

        Rotates the shape around the z-axis the amount specified by the ``angle``
        parameter. Angles should be specified in radians (values from 0 to TWO_PI) or
        converted from degrees to radians with the ``radians()`` method.

        Shapes are always rotated around the upper-left corner of their bounding box.
        Positive numbers rotate objects in a clockwise direction. Subsequent calls to
        the method accumulates the effect. For example, calling ``rotate_z(HALF_PI)``
        and then ``rotate_z(HALF_PI)`` is the same as ``rotate_z(PI)``. This
        transformation is applied directly to the shape, it's not refreshed each time
        ``draw()`` is run.

        This method requires a 3D renderer. You need to use P3D as a third parameter for
        the ``size()`` function as shown in the example above.
"""
        return self._instance.rotateZ(angle)

    @overload
    def scale(self, s: float, /) -> None:
        """Increases or decreases the size of a shape by expanding and contracting
        vertices.

        Underlying Java method: PShape.scale

        Methods
        -------

        You can use any of the following signatures:

         * scale(s: float) -> None
         * scale(x: float, y: float) -> None
         * scale(x: float, y: float, z: float) -> None

        Parameters
        ----------

        s: float
            percentate to scale the object

        x: float
            percentage to scale the object in the x-axis

        y: float
            percentage to scale the object in the y-axis

        z: float
            percentage to scale the object in the z-axis

        Notes
        -----

        Increases or decreases the size of a shape by expanding and contracting
        vertices. Shapes always scale from the relative origin of their bounding box.
        Scale values are specified as decimal percentages. For example, the method call
        ``scale(2.0)`` increases the dimension of a shape by 200%. Subsequent calls to
        the method multiply the effect. For example, calling ``scale(2.0)`` and then
        ``scale(1.5)`` is the same as ``scale(3.0)``. This transformation is applied
        directly to the shape; it's not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        pass

    @overload
    def scale(self, x: float, y: float, /) -> None:
        """Increases or decreases the size of a shape by expanding and contracting
        vertices.

        Underlying Java method: PShape.scale

        Methods
        -------

        You can use any of the following signatures:

         * scale(s: float) -> None
         * scale(x: float, y: float) -> None
         * scale(x: float, y: float, z: float) -> None

        Parameters
        ----------

        s: float
            percentate to scale the object

        x: float
            percentage to scale the object in the x-axis

        y: float
            percentage to scale the object in the y-axis

        z: float
            percentage to scale the object in the z-axis

        Notes
        -----

        Increases or decreases the size of a shape by expanding and contracting
        vertices. Shapes always scale from the relative origin of their bounding box.
        Scale values are specified as decimal percentages. For example, the method call
        ``scale(2.0)`` increases the dimension of a shape by 200%. Subsequent calls to
        the method multiply the effect. For example, calling ``scale(2.0)`` and then
        ``scale(1.5)`` is the same as ``scale(3.0)``. This transformation is applied
        directly to the shape; it's not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        pass

    @overload
    def scale(self, x: float, y: float, z: float, /) -> None:
        """Increases or decreases the size of a shape by expanding and contracting
        vertices.

        Underlying Java method: PShape.scale

        Methods
        -------

        You can use any of the following signatures:

         * scale(s: float) -> None
         * scale(x: float, y: float) -> None
         * scale(x: float, y: float, z: float) -> None

        Parameters
        ----------

        s: float
            percentate to scale the object

        x: float
            percentage to scale the object in the x-axis

        y: float
            percentage to scale the object in the y-axis

        z: float
            percentage to scale the object in the z-axis

        Notes
        -----

        Increases or decreases the size of a shape by expanding and contracting
        vertices. Shapes always scale from the relative origin of their bounding box.
        Scale values are specified as decimal percentages. For example, the method call
        ``scale(2.0)`` increases the dimension of a shape by 200%. Subsequent calls to
        the method multiply the effect. For example, calling ``scale(2.0)`` and then
        ``scale(1.5)`` is the same as ``scale(3.0)``. This transformation is applied
        directly to the shape; it's not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        pass

    def scale(self, *args):
        """Increases or decreases the size of a shape by expanding and contracting
        vertices.

        Underlying Java method: PShape.scale

        Methods
        -------

        You can use any of the following signatures:

         * scale(s: float) -> None
         * scale(x: float, y: float) -> None
         * scale(x: float, y: float, z: float) -> None

        Parameters
        ----------

        s: float
            percentate to scale the object

        x: float
            percentage to scale the object in the x-axis

        y: float
            percentage to scale the object in the y-axis

        z: float
            percentage to scale the object in the z-axis

        Notes
        -----

        Increases or decreases the size of a shape by expanding and contracting
        vertices. Shapes always scale from the relative origin of their bounding box.
        Scale values are specified as decimal percentages. For example, the method call
        ``scale(2.0)`` increases the dimension of a shape by 200%. Subsequent calls to
        the method multiply the effect. For example, calling ``scale(2.0)`` and then
        ``scale(1.5)`` is the same as ``scale(3.0)``. This transformation is applied
        directly to the shape; it's not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        return self._instance.scale(*args)

    @overload
    def set_ambient(self, ambient: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setAmbient

        Methods
        -------

        You can use any of the following signatures:

         * set_ambient(ambient: int) -> None
         * set_ambient(index: int, ambient: int) -> None

        Parameters
        ----------

        ambient: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_ambient(self, index: int, ambient: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setAmbient

        Methods
        -------

        You can use any of the following signatures:

         * set_ambient(ambient: int) -> None
         * set_ambient(index: int, ambient: int) -> None

        Parameters
        ----------

        ambient: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def set_ambient(self, *args):
        """new template no description.

        Underlying Java method: PShape.setAmbient

        Methods
        -------

        You can use any of the following signatures:

         * set_ambient(ambient: int) -> None
         * set_ambient(index: int, ambient: int) -> None

        Parameters
        ----------

        ambient: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setAmbient(*args)

    @overload
    def set_attrib(self, name: str, index: int, /, *values: bool) -> None:
        """new template no description.

        Underlying Java method: PShape.setAttrib

        Methods
        -------

        You can use any of the following signatures:

         * set_attrib(name: str, index: int, values: bool) -> None
         * set_attrib(name: str, index: int, values: float) -> None
         * set_attrib(name: str, index: int, values: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_attrib(self, name: str, index: int, /, *values: float) -> None:
        """new template no description.

        Underlying Java method: PShape.setAttrib

        Methods
        -------

        You can use any of the following signatures:

         * set_attrib(name: str, index: int, values: bool) -> None
         * set_attrib(name: str, index: int, values: float) -> None
         * set_attrib(name: str, index: int, values: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_attrib(self, name: str, index: int, /, *values: int) -> None:
        """new template no description.

        Underlying Java method: PShape.setAttrib

        Methods
        -------

        You can use any of the following signatures:

         * set_attrib(name: str, index: int, values: bool) -> None
         * set_attrib(name: str, index: int, values: float) -> None
         * set_attrib(name: str, index: int, values: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @_py5shape_type_fixer
    def set_attrib(self, *args):
        """new template no description.

        Underlying Java method: PShape.setAttrib

        Methods
        -------

        You can use any of the following signatures:

         * set_attrib(name: str, index: int, values: bool) -> None
         * set_attrib(name: str, index: int, values: float) -> None
         * set_attrib(name: str, index: int, values: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        name: str
            missing variable description

        values: bool
            missing variable description

        values: float
            missing variable description

        values: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setAttrib(*args)

    @overload
    def set_emissive(self, emissive: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setEmissive

        Methods
        -------

        You can use any of the following signatures:

         * set_emissive(emissive: int) -> None
         * set_emissive(index: int, emissive: int) -> None

        Parameters
        ----------

        emissive: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_emissive(self, index: int, emissive: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setEmissive

        Methods
        -------

        You can use any of the following signatures:

         * set_emissive(emissive: int) -> None
         * set_emissive(index: int, emissive: int) -> None

        Parameters
        ----------

        emissive: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def set_emissive(self, *args):
        """new template no description.

        Underlying Java method: PShape.setEmissive

        Methods
        -------

        You can use any of the following signatures:

         * set_emissive(emissive: int) -> None
         * set_emissive(index: int, emissive: int) -> None

        Parameters
        ----------

        emissive: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setEmissive(*args)

    @overload
    def set_fill(self, fill: bool, /) -> None:
        """The ``set_fill()`` method defines the fill color of a ``Py5Shape``.

        Underlying Java method: PShape.setFill

        Methods
        -------

        You can use any of the following signatures:

         * set_fill(fill: bool) -> None
         * set_fill(fill: int) -> None
         * set_fill(index: int, fill: int) -> None

        Parameters
        ----------

        fill: bool
            missing variable description

        fill: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        The ``set_fill()`` method defines the fill color of a ``Py5Shape``. This method
        is used after shapes are created or when a shape is defined explicitly (e.g.
        ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example. When a
        shape is created with ``begin_shape()`` and ``end_shape()``, its attributes may
        be changed with ``fill()`` and ``stroke()`` within ``begin_shape()`` and
        ``end_shape()``. However, after the shape is created, only the ``set_fill()``
        method can define a new fill value for the ``Py5Shape``.
"""
        pass

    @overload
    def set_fill(self, fill: int, /) -> None:
        """The ``set_fill()`` method defines the fill color of a ``Py5Shape``.

        Underlying Java method: PShape.setFill

        Methods
        -------

        You can use any of the following signatures:

         * set_fill(fill: bool) -> None
         * set_fill(fill: int) -> None
         * set_fill(index: int, fill: int) -> None

        Parameters
        ----------

        fill: bool
            missing variable description

        fill: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        The ``set_fill()`` method defines the fill color of a ``Py5Shape``. This method
        is used after shapes are created or when a shape is defined explicitly (e.g.
        ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example. When a
        shape is created with ``begin_shape()`` and ``end_shape()``, its attributes may
        be changed with ``fill()`` and ``stroke()`` within ``begin_shape()`` and
        ``end_shape()``. However, after the shape is created, only the ``set_fill()``
        method can define a new fill value for the ``Py5Shape``.
"""
        pass

    @overload
    def set_fill(self, index: int, fill: int, /) -> None:
        """The ``set_fill()`` method defines the fill color of a ``Py5Shape``.

        Underlying Java method: PShape.setFill

        Methods
        -------

        You can use any of the following signatures:

         * set_fill(fill: bool) -> None
         * set_fill(fill: int) -> None
         * set_fill(index: int, fill: int) -> None

        Parameters
        ----------

        fill: bool
            missing variable description

        fill: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        The ``set_fill()`` method defines the fill color of a ``Py5Shape``. This method
        is used after shapes are created or when a shape is defined explicitly (e.g.
        ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example. When a
        shape is created with ``begin_shape()`` and ``end_shape()``, its attributes may
        be changed with ``fill()`` and ``stroke()`` within ``begin_shape()`` and
        ``end_shape()``. However, after the shape is created, only the ``set_fill()``
        method can define a new fill value for the ``Py5Shape``.
"""
        pass

    @_py5shape_type_fixer
    def set_fill(self, *args):
        """The ``set_fill()`` method defines the fill color of a ``Py5Shape``.

        Underlying Java method: PShape.setFill

        Methods
        -------

        You can use any of the following signatures:

         * set_fill(fill: bool) -> None
         * set_fill(fill: int) -> None
         * set_fill(index: int, fill: int) -> None

        Parameters
        ----------

        fill: bool
            missing variable description

        fill: int
            missing variable description

        index: int
            missing variable description

        Notes
        -----

        The ``set_fill()`` method defines the fill color of a ``Py5Shape``. This method
        is used after shapes are created or when a shape is defined explicitly (e.g.
        ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example. When a
        shape is created with ``begin_shape()`` and ``end_shape()``, its attributes may
        be changed with ``fill()`` and ``stroke()`` within ``begin_shape()`` and
        ``end_shape()``. However, after the shape is created, only the ``set_fill()``
        method can define a new fill value for the ``Py5Shape``.
"""
        return self._instance.setFill(*args)

    def set_name(self, name: str, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setName

        Parameters
        ----------

        name: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setName(name)

    def set_normal(self, index: int, nx: float,
                   ny: float, nz: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setNormal

        Parameters
        ----------

        index: int
            missing variable description

        nx: float
            missing variable description

        ny: float
            missing variable description

        nz: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setNormal(index, nx, ny, nz)

    @overload
    def set_shininess(self, shine: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setShininess

        Methods
        -------

        You can use any of the following signatures:

         * set_shininess(index: int, shine: float) -> None
         * set_shininess(shine: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        shine: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_shininess(self, index: int, shine: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setShininess

        Methods
        -------

        You can use any of the following signatures:

         * set_shininess(index: int, shine: float) -> None
         * set_shininess(shine: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        shine: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def set_shininess(self, *args):
        """new template no description.

        Underlying Java method: PShape.setShininess

        Methods
        -------

        You can use any of the following signatures:

         * set_shininess(index: int, shine: float) -> None
         * set_shininess(shine: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        shine: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setShininess(*args)

    @overload
    def set_specular(self, specular: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setSpecular

        Methods
        -------

        You can use any of the following signatures:

         * set_specular(index: int, specular: int) -> None
         * set_specular(specular: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        specular: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_specular(self, index: int, specular: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setSpecular

        Methods
        -------

        You can use any of the following signatures:

         * set_specular(index: int, specular: int) -> None
         * set_specular(specular: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        specular: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def set_specular(self, *args):
        """new template no description.

        Underlying Java method: PShape.setSpecular

        Methods
        -------

        You can use any of the following signatures:

         * set_specular(index: int, specular: int) -> None
         * set_specular(specular: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        specular: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setSpecular(*args)

    @overload
    def set_stroke(self, stroke: bool, /) -> None:
        """The ``set_stroke()`` method defines the outline color of a ``Py5Shape``.

        Underlying Java method: PShape.setStroke

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke(index: int, stroke: int) -> None
         * set_stroke(stroke: bool) -> None
         * set_stroke(stroke: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        stroke: bool
            missing variable description

        stroke: int
            missing variable description

        Notes
        -----

        The ``set_stroke()`` method defines the outline color of a ``Py5Shape``. This
        method is used after shapes are created or when a shape is defined explicitly
        (e.g. ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example.
        When a shape is created with ``begin_shape()`` and ``end_shape()``, its
        attributes may be changed with ``fill()`` and ``stroke()`` within
        ``begin_shape()`` and ``end_shape()``. However, after the shape is created, only
        the ``set_stroke()`` method can define a new stroke value for the ``Py5Shape``.
"""
        pass

    @overload
    def set_stroke(self, stroke: int, /) -> None:
        """The ``set_stroke()`` method defines the outline color of a ``Py5Shape``.

        Underlying Java method: PShape.setStroke

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke(index: int, stroke: int) -> None
         * set_stroke(stroke: bool) -> None
         * set_stroke(stroke: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        stroke: bool
            missing variable description

        stroke: int
            missing variable description

        Notes
        -----

        The ``set_stroke()`` method defines the outline color of a ``Py5Shape``. This
        method is used after shapes are created or when a shape is defined explicitly
        (e.g. ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example.
        When a shape is created with ``begin_shape()`` and ``end_shape()``, its
        attributes may be changed with ``fill()`` and ``stroke()`` within
        ``begin_shape()`` and ``end_shape()``. However, after the shape is created, only
        the ``set_stroke()`` method can define a new stroke value for the ``Py5Shape``.
"""
        pass

    @overload
    def set_stroke(self, index: int, stroke: int, /) -> None:
        """The ``set_stroke()`` method defines the outline color of a ``Py5Shape``.

        Underlying Java method: PShape.setStroke

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke(index: int, stroke: int) -> None
         * set_stroke(stroke: bool) -> None
         * set_stroke(stroke: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        stroke: bool
            missing variable description

        stroke: int
            missing variable description

        Notes
        -----

        The ``set_stroke()`` method defines the outline color of a ``Py5Shape``. This
        method is used after shapes are created or when a shape is defined explicitly
        (e.g. ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example.
        When a shape is created with ``begin_shape()`` and ``end_shape()``, its
        attributes may be changed with ``fill()`` and ``stroke()`` within
        ``begin_shape()`` and ``end_shape()``. However, after the shape is created, only
        the ``set_stroke()`` method can define a new stroke value for the ``Py5Shape``.
"""
        pass

    @_py5shape_type_fixer
    def set_stroke(self, *args):
        """The ``set_stroke()`` method defines the outline color of a ``Py5Shape``.

        Underlying Java method: PShape.setStroke

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke(index: int, stroke: int) -> None
         * set_stroke(stroke: bool) -> None
         * set_stroke(stroke: int) -> None

        Parameters
        ----------

        index: int
            missing variable description

        stroke: bool
            missing variable description

        stroke: int
            missing variable description

        Notes
        -----

        The ``set_stroke()`` method defines the outline color of a ``Py5Shape``. This
        method is used after shapes are created or when a shape is defined explicitly
        (e.g. ``create_shape(RECT, 20, 20, 80, 80)``) as shown in the above example.
        When a shape is created with ``begin_shape()`` and ``end_shape()``, its
        attributes may be changed with ``fill()`` and ``stroke()`` within
        ``begin_shape()`` and ``end_shape()``. However, after the shape is created, only
        the ``set_stroke()`` method can define a new stroke value for the ``Py5Shape``.
"""
        return self._instance.setStroke(*args)

    def set_stroke_cap(self, cap: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setStrokeCap

        Parameters
        ----------

        cap: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setStrokeCap(cap)

    def set_stroke_join(self, join: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setStrokeJoin

        Parameters
        ----------

        join: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setStrokeJoin(join)

    @overload
    def set_stroke_weight(self, weight: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setStrokeWeight

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke_weight(index: int, weight: float) -> None
         * set_stroke_weight(weight: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        weight: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_stroke_weight(self, index: int, weight: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setStrokeWeight

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke_weight(index: int, weight: float) -> None
         * set_stroke_weight(weight: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        weight: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    def set_stroke_weight(self, *args):
        """new template no description.

        Underlying Java method: PShape.setStrokeWeight

        Methods
        -------

        You can use any of the following signatures:

         * set_stroke_weight(index: int, weight: float) -> None
         * set_stroke_weight(weight: float) -> None

        Parameters
        ----------

        index: int
            missing variable description

        weight: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setStrokeWeight(*args)

    def set_texture(self, tex: Py5Image, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTexture

        Parameters
        ----------

        tex: Py5Image
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setTexture(tex)

    def set_texture_mode(self, mode: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTextureMode

        Parameters
        ----------

        mode: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setTextureMode(mode)

    def set_texture_uv(self, index: int, u: float, v: float, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTextureUV

        Parameters
        ----------

        index: int
            missing variable description

        u: float
            missing variable description

        v: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setTextureUV(index, u, v)

    @overload
    def set_tint(self, tint: bool, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTint

        Methods
        -------

        You can use any of the following signatures:

         * set_tint(fill: int) -> None
         * set_tint(index: int, tint: int) -> None
         * set_tint(tint: bool) -> None

        Parameters
        ----------

        fill: int
            missing variable description

        index: int
            missing variable description

        tint: bool
            missing variable description

        tint: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_tint(self, fill: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTint

        Methods
        -------

        You can use any of the following signatures:

         * set_tint(fill: int) -> None
         * set_tint(index: int, tint: int) -> None
         * set_tint(tint: bool) -> None

        Parameters
        ----------

        fill: int
            missing variable description

        index: int
            missing variable description

        tint: bool
            missing variable description

        tint: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def set_tint(self, index: int, tint: int, /) -> None:
        """new template no description.

        Underlying Java method: PShape.setTint

        Methods
        -------

        You can use any of the following signatures:

         * set_tint(fill: int) -> None
         * set_tint(index: int, tint: int) -> None
         * set_tint(tint: bool) -> None

        Parameters
        ----------

        fill: int
            missing variable description

        index: int
            missing variable description

        tint: bool
            missing variable description

        tint: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @_py5shape_type_fixer
    def set_tint(self, *args):
        """new template no description.

        Underlying Java method: PShape.setTint

        Methods
        -------

        You can use any of the following signatures:

         * set_tint(fill: int) -> None
         * set_tint(index: int, tint: int) -> None
         * set_tint(tint: bool) -> None

        Parameters
        ----------

        fill: int
            missing variable description

        index: int
            missing variable description

        tint: bool
            missing variable description

        tint: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setTint(*args)

    @overload
    def set_vertex(self, index: int, x: float, y: float, /) -> None:
        """The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.setVertex

        Methods
        -------

        You can use any of the following signatures:

         * set_vertex(index: int, vec: NDArray[(Any,), Float]) -> None
         * set_vertex(index: int, x: float, y: float) -> None
         * set_vertex(index: int, x: float, y: float, z: float) -> None

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            the PVector to define the x, y, z coordinates

        x: float
            the x value for the vertex

        y: float
            the y value for the vertex

        z: float
            the z value for the vertex

        Notes
        -----

        The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter. This method works when
        shapes are created as shown in the example above, but won't work properly when a
        shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20, 80, 80)``.
"""
        pass

    @overload
    def set_vertex(self, index: int, x: float, y: float, z: float, /) -> None:
        """The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.setVertex

        Methods
        -------

        You can use any of the following signatures:

         * set_vertex(index: int, vec: NDArray[(Any,), Float]) -> None
         * set_vertex(index: int, x: float, y: float) -> None
         * set_vertex(index: int, x: float, y: float, z: float) -> None

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            the PVector to define the x, y, z coordinates

        x: float
            the x value for the vertex

        y: float
            the y value for the vertex

        z: float
            the z value for the vertex

        Notes
        -----

        The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter. This method works when
        shapes are created as shown in the example above, but won't work properly when a
        shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20, 80, 80)``.
"""
        pass

    @overload
    def set_vertex(self, index: int, vec: NDArray[(Any,), Float], /) -> None:
        """The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.setVertex

        Methods
        -------

        You can use any of the following signatures:

         * set_vertex(index: int, vec: NDArray[(Any,), Float]) -> None
         * set_vertex(index: int, x: float, y: float) -> None
         * set_vertex(index: int, x: float, y: float, z: float) -> None

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            the PVector to define the x, y, z coordinates

        x: float
            the x value for the vertex

        y: float
            the y value for the vertex

        z: float
            the z value for the vertex

        Notes
        -----

        The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter. This method works when
        shapes are created as shown in the example above, but won't work properly when a
        shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20, 80, 80)``.
"""
        pass

    def set_vertex(self, *args):
        """The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter.

        Underlying Java method: PShape.setVertex

        Methods
        -------

        You can use any of the following signatures:

         * set_vertex(index: int, vec: NDArray[(Any,), Float]) -> None
         * set_vertex(index: int, x: float, y: float) -> None
         * set_vertex(index: int, x: float, y: float, z: float) -> None

        Parameters
        ----------

        index: int
            the location of the vertex

        vec: NDArray[(Any,), Float]
            the PVector to define the x, y, z coordinates

        x: float
            the x value for the vertex

        y: float
            the y value for the vertex

        z: float
            the z value for the vertex

        Notes
        -----

        The ``set_vertex()`` method defines the coordinates of the vertex point located
        at the position defined by the ``index`` parameter. This method works when
        shapes are created as shown in the example above, but won't work properly when a
        shape is defined explicitly (e.g. ``create_shape(RECT, 20, 20, 80, 80)``.
"""
        return self._instance.setVertex(*args)

    def set_visible(self, visible: bool, /) -> None:
        """Sets the shape to be visible or invisible.

        Underlying Java method: PShape.setVisible

        Parameters
        ----------

        visible: bool
            "false" makes the shape invisible and "true" makes it visible

        Notes
        -----

        Sets the shape to be visible or invisible. This is determined by the value of
        the ``visible`` parameter.

        The default visibility of a shape is usually controlled by whatever program
        created the SVG file. For instance, this parameter is controlled by showing or
        hiding the shape in the layers palette in Adobe Illustrator.
"""
        return self._instance.setVisible(visible)

    @overload
    def translate(self, x: float, y: float, /) -> None:
        """Specifies an amount to displace the shape.

        Underlying Java method: PShape.translate

        Methods
        -------

        You can use any of the following signatures:

         * translate(x: float, y: float) -> None
         * translate(x: float, y: float, z: float) -> None

        Parameters
        ----------

        x: float
            left/right translation

        y: float
            up/down translation

        z: float
            forward/back translation

        Notes
        -----

        Specifies an amount to displace the shape. The ``x`` parameter specifies
        left/right translation, the ``y`` parameter specifies up/down translation, and
        the ``z`` parameter specifies translations toward/away from the screen.
        Subsequent calls to the method accumulates the effect. For example, calling
        ``translate(50, 0)`` and then ``translate(20, 0)`` is the same as
        ``translate(70, 0)``. This transformation is applied directly to the shape, it's
        not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        pass

    @overload
    def translate(self, x: float, y: float, z: float, /) -> None:
        """Specifies an amount to displace the shape.

        Underlying Java method: PShape.translate

        Methods
        -------

        You can use any of the following signatures:

         * translate(x: float, y: float) -> None
         * translate(x: float, y: float, z: float) -> None

        Parameters
        ----------

        x: float
            left/right translation

        y: float
            up/down translation

        z: float
            forward/back translation

        Notes
        -----

        Specifies an amount to displace the shape. The ``x`` parameter specifies
        left/right translation, the ``y`` parameter specifies up/down translation, and
        the ``z`` parameter specifies translations toward/away from the screen.
        Subsequent calls to the method accumulates the effect. For example, calling
        ``translate(50, 0)`` and then ``translate(20, 0)`` is the same as
        ``translate(70, 0)``. This transformation is applied directly to the shape, it's
        not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        pass

    def translate(self, *args):
        """Specifies an amount to displace the shape.

        Underlying Java method: PShape.translate

        Methods
        -------

        You can use any of the following signatures:

         * translate(x: float, y: float) -> None
         * translate(x: float, y: float, z: float) -> None

        Parameters
        ----------

        x: float
            left/right translation

        y: float
            up/down translation

        z: float
            forward/back translation

        Notes
        -----

        Specifies an amount to displace the shape. The ``x`` parameter specifies
        left/right translation, the ``y`` parameter specifies up/down translation, and
        the ``z`` parameter specifies translations toward/away from the screen.
        Subsequent calls to the method accumulates the effect. For example, calling
        ``translate(50, 0)`` and then ``translate(20, 0)`` is the same as
        ``translate(70, 0)``. This transformation is applied directly to the shape, it's
        not refreshed each time ``draw()`` is run.

        Using this method with the ``z`` parameter requires using the P3D parameter in
        combination with size.
"""
        return self._instance.translate(*args)



# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***
from __future__ import annotations

import functools
from typing import overload, List, Union  # noqa

from .base import Py5Base
from .mixins import PixelMixin


def _return_py5image(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        ret = f(self_, *args)
        if ret is None or isinstance(ret, int):
            return ret
        else:
            return Py5Image(ret)
    return decorated


class Py5Image(PixelMixin, Py5Base):
    """Datatype for storing images.

    Underlying Java class: PImage.PImage

    Notes
    -----

    Datatype for storing images. Processing can display ``.gif``, ``.jpg``,
    ``.tga``, and ``.png`` images. Images may be displayed in 2D and 3D space.
    Before an image is used, it must be loaded with the ``load_image()`` function.
    The ``Py5Image`` class contains fields for the ``width`` and ``height`` of the
    image, as well as an array called ``pixels[]`` that contains the values for
    every pixel in the image. The methods described below allow easy access to the
    image's pixels and alpha channel and simplify the process of compositing.

    Before using the ``pixels[]`` array, be sure to use the ``load_pixels()`` method
    on the image to make sure that the pixel data is properly loaded.

    To create a new image, use the ``create_image()`` function. Do not use the
    syntax ``new Py5Image()``.
"""

    def __init__(self, pimage):
        self._instance = pimage
        super().__init__(instance=pimage)

    ADD = 2
    ALPHA = 4
    ALPHA_MASK = -16777216
    ARGB = 2
    BLEND = 1
    BLUE_MASK = 255
    BLUR = 11
    BURN = 8192
    DARKEST = 16
    DIFFERENCE = 32
    DILATE = 18
    DODGE = 4096
    ERODE = 17
    EXCLUSION = 64
    GIF = 3
    GRAY = 12
    GREEN_MASK = 65280
    HARD_LIGHT = 1024
    HSB = 3
    INVERT = 13
    JPEG = 2
    LIGHTEST = 8
    MULTIPLY = 128
    OPAQUE = 14
    OVERLAY = 512
    POSTERIZE = 15
    RED_MASK = 16711680
    REPLACE = 0
    RGB = 1
    SCREEN = 256
    SOFT_LIGHT = 2048
    SUBTRACT = 4
    TARGA = 1
    THRESHOLD = 16
    TIFF = 0

    def _get_height(self) -> int:
        """The height of the image in units of pixels.

        Underlying Java field: PImage.height

        Notes
        -----

        The height of the image in units of pixels.
"""
        return self._instance.height
    height: int = property(fget=_get_height)

    def _get_pixel_density(self) -> int:
        """This function is new with Processing 3.0.

        Underlying Java method: PApplet.pixelDensity

        Parameters
        ----------

        density: int
            1 or 2

        Notes
        -----

        This function is new with Processing 3.0. It makes it possible for Processing to
        render using all of the pixels on high resolutions screens like Apple Retina
        displays and Windows High-DPI displays. This function can only be run once
        within a program and it must be used right after ``size()`` in a program without
        a ``setup()`` and used within ``setup()`` when a program has one.  The
        ``pixel_density()`` should only be used with hardcoded numbers (in almost all
        cases this number will be 2) or in combination with ``display_density()`` as in
        the third example above.

        When the pixel density is set to more than 1, it changes all of the pixel
        operations including the way ``get()``, ``set()``, ``blend()``, ``copy()``, and
        ``update_pixels()`` all work. See the reference for ``pixel_width`` and
        ``pixel_height`` for more information.

        To use variables as the arguments to ``pixel_density()`` function, place the
        ``pixel_density()`` function within the ``settings()`` function. There is more
        information about this on the ``settings()`` reference page.
"""
        return self._instance.pixelDensity
    pixel_density: int = property(fget=_get_pixel_density)

    def _get_pixel_height(self) -> int:
        """When ``pixel_density(2)`` is used to make use of a high resolution display
        (called a Retina display on OS X or high-dpi on Windows and Linux), the width
        and height of the sketch do not change, but the number of pixels is doubled.

        Underlying Java field: PApplet.pixelHeight

        Notes
        -----

        When ``pixel_density(2)`` is used to make use of a high resolution display
        (called a Retina display on OS X or high-dpi on Windows and Linux), the width
        and height of the sketch do not change, but the number of pixels is doubled. As
        a result, all operations that use pixels (like ``load_pixels()``, ``get()``,
        ``set()``, etc.) happen in this doubled space. As a convenience, the variables
        ``pixel_width`` and ``pixel_height`` hold the actual width and height of the
        sketch in pixels. This is useful for any sketch that uses the ``pixels[]``
        array, for instance, because the number of elements in the array will be
        ``pixel_width*pixel_height``, not ``width*height``.
"""
        return self._instance.pixelHeight
    pixel_height: int = property(fget=_get_pixel_height)

    def _get_pixel_width(self) -> int:
        """When ``pixel_density(2)`` is used to make use of a high resolution display
        (called a Retina display on OS X or high-dpi on Windows and Linux), the width
        and height of the sketch do not change, but the number of pixels is doubled.

        Underlying Java field: PApplet.pixelWidth

        Notes
        -----

        When ``pixel_density(2)`` is used to make use of a high resolution display
        (called a Retina display on OS X or high-dpi on Windows and Linux), the width
        and height of the sketch do not change, but the number of pixels is doubled. As
        a result, all operations that use pixels (like ``load_pixels()``, ``get()``,
        ``set()``, etc.) happen in this doubled space. As a convenience, the variables
        ``pixel_width`` and ``pixel_height`` hold the actual width and height of the
        sketch in pixels. This is useful for any sketch that uses the ``pixels[]``
        array, for instance, because the number of elements in the array will be
        ``pixel_width*pixel_height``, not ``width*height``.
"""
        return self._instance.pixelWidth
    pixel_width: int = property(fget=_get_pixel_width)

    def _get_pixels(self) -> JArray(JInt):
        """The pixels[] array contains the values for all the pixels in the image.

        Underlying Java field: PImage.pixels

        Notes
        -----

        The pixels[] array contains the values for all the pixels in the image. These
        values are of the color datatype. This array is the size of the image, meaning
        if the image is 100 x 100 pixels, there will be 10,000 values and if the window
        is 200 x 300 pixels, there will be 60,000 values.

        Before accessing this array, the data must loaded with the ``load_pixels()``
        method. Failure to do so may result in a NullPointerException. After the array
        data has been modified, the ``update_pixels()`` method must be run to update the
        content of the display window.
"""
        return self._instance.pixels
    pixels: JArray(JInt) = property(fget=_get_pixels)

    def _get_width(self) -> int:
        """The width of the image in units of pixels.

        Underlying Java field: PImage.width

        Notes
        -----

        The width of the image in units of pixels.
"""
        return self._instance.width
    width: int = property(fget=_get_width)

    @overload
    def blend(self, sx: int, sy: int, sw: int, sh: int, dx: int,
              dy: int, dw: int, dh: int, mode: int, /) -> None:
        """Blends a region of pixels into the image specified by the ``img`` parameter.

        Underlying Java method: PImage.blend

        Methods
        -------

        You can use any of the following signatures:

         * blend(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None
         * blend(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destinations's upper left corner

        dy: int
            Y coordinate of the destinations's upper left corner

        mode: int
            Either BLEND, ADD, SUBTRACT, LIGHTEST, DARKEST, DIFFERENCE, EXCLUSION, MULTIPLY, SCREEN, OVERLAY, HARD_LIGHT, SOFT_LIGHT, DODGE, BURN

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Blends a region of pixels into the image specified by the ``img`` parameter.
        These copies utilize full alpha channel support and a choice of the following
        modes to blend the colors of source pixels (A) with the ones of pixels in the
        destination image (B):

        BLEND - linear interpolation of colours: C = A*factor + B

        ADD - additive blending with white clip: C = min(A*factor + B, 255)

        SUBTRACT - subtractive blending with black clip: C = max(B - A*factor, 0)

        DARKEST - only the darkest colour succeeds: C = min(A*factor, B)

        LIGHTEST - only the lightest colour succeeds: C = max(A*factor, B)

        DIFFERENCE - subtract colors from underlying image.

        EXCLUSION - similar to DIFFERENCE, but less extreme.

        MULTIPLY - Multiply the colors, result will always be darker.

        SCREEN - Opposite multiply, uses inverse values of the colors.

        OVERLAY - A mix of MULTIPLY and SCREEN. Multiplies dark values,
        and screens light values.

        HARD_LIGHT - SCREEN when greater than 50% gray, MULTIPLY when lower.

        SOFT_LIGHT - Mix of DARKEST and LIGHTEST.
        Works like OVERLAY, but not as harsh.

        DODGE - Lightens light tones and increases contrast, ignores darks.
        Called "Color Dodge" in Illustrator and Photoshop.

        BURN - Darker areas are applied, increasing contrast, ignores lights.
        Called "Color Burn" in Illustrator and Photoshop.

        All modes use the alpha information (highest byte) of source image pixels as the
        blending factor. If the source and destination regions are different sizes, the
        image will be automatically resized to match the destination size. If the
        ``src`` parameter is not used, the display window is used as the source image.

        As of release 0149, this function ignores ``image_mode()``.
"""
        pass

    @overload
    def blend(self, src: Py5Image, sx: int, sy: int, sw: int, sh: int,
              dx: int, dy: int, dw: int, dh: int, mode: int, /) -> None:
        """Blends a region of pixels into the image specified by the ``img`` parameter.

        Underlying Java method: PImage.blend

        Methods
        -------

        You can use any of the following signatures:

         * blend(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None
         * blend(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destinations's upper left corner

        dy: int
            Y coordinate of the destinations's upper left corner

        mode: int
            Either BLEND, ADD, SUBTRACT, LIGHTEST, DARKEST, DIFFERENCE, EXCLUSION, MULTIPLY, SCREEN, OVERLAY, HARD_LIGHT, SOFT_LIGHT, DODGE, BURN

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Blends a region of pixels into the image specified by the ``img`` parameter.
        These copies utilize full alpha channel support and a choice of the following
        modes to blend the colors of source pixels (A) with the ones of pixels in the
        destination image (B):

        BLEND - linear interpolation of colours: C = A*factor + B

        ADD - additive blending with white clip: C = min(A*factor + B, 255)

        SUBTRACT - subtractive blending with black clip: C = max(B - A*factor, 0)

        DARKEST - only the darkest colour succeeds: C = min(A*factor, B)

        LIGHTEST - only the lightest colour succeeds: C = max(A*factor, B)

        DIFFERENCE - subtract colors from underlying image.

        EXCLUSION - similar to DIFFERENCE, but less extreme.

        MULTIPLY - Multiply the colors, result will always be darker.

        SCREEN - Opposite multiply, uses inverse values of the colors.

        OVERLAY - A mix of MULTIPLY and SCREEN. Multiplies dark values,
        and screens light values.

        HARD_LIGHT - SCREEN when greater than 50% gray, MULTIPLY when lower.

        SOFT_LIGHT - Mix of DARKEST and LIGHTEST.
        Works like OVERLAY, but not as harsh.

        DODGE - Lightens light tones and increases contrast, ignores darks.
        Called "Color Dodge" in Illustrator and Photoshop.

        BURN - Darker areas are applied, increasing contrast, ignores lights.
        Called "Color Burn" in Illustrator and Photoshop.

        All modes use the alpha information (highest byte) of source image pixels as the
        blending factor. If the source and destination regions are different sizes, the
        image will be automatically resized to match the destination size. If the
        ``src`` parameter is not used, the display window is used as the source image.

        As of release 0149, this function ignores ``image_mode()``.
"""
        pass

    def blend(self, *args):
        """Blends a region of pixels into the image specified by the ``img`` parameter.

        Underlying Java method: PImage.blend

        Methods
        -------

        You can use any of the following signatures:

         * blend(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None
         * blend(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int, mode: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destinations's upper left corner

        dy: int
            Y coordinate of the destinations's upper left corner

        mode: int
            Either BLEND, ADD, SUBTRACT, LIGHTEST, DARKEST, DIFFERENCE, EXCLUSION, MULTIPLY, SCREEN, OVERLAY, HARD_LIGHT, SOFT_LIGHT, DODGE, BURN

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Blends a region of pixels into the image specified by the ``img`` parameter.
        These copies utilize full alpha channel support and a choice of the following
        modes to blend the colors of source pixels (A) with the ones of pixels in the
        destination image (B):

        BLEND - linear interpolation of colours: C = A*factor + B

        ADD - additive blending with white clip: C = min(A*factor + B, 255)

        SUBTRACT - subtractive blending with black clip: C = max(B - A*factor, 0)

        DARKEST - only the darkest colour succeeds: C = min(A*factor, B)

        LIGHTEST - only the lightest colour succeeds: C = max(A*factor, B)

        DIFFERENCE - subtract colors from underlying image.

        EXCLUSION - similar to DIFFERENCE, but less extreme.

        MULTIPLY - Multiply the colors, result will always be darker.

        SCREEN - Opposite multiply, uses inverse values of the colors.

        OVERLAY - A mix of MULTIPLY and SCREEN. Multiplies dark values,
        and screens light values.

        HARD_LIGHT - SCREEN when greater than 50% gray, MULTIPLY when lower.

        SOFT_LIGHT - Mix of DARKEST and LIGHTEST.
        Works like OVERLAY, but not as harsh.

        DODGE - Lightens light tones and increases contrast, ignores darks.
        Called "Color Dodge" in Illustrator and Photoshop.

        BURN - Darker areas are applied, increasing contrast, ignores lights.
        Called "Color Burn" in Illustrator and Photoshop.

        All modes use the alpha information (highest byte) of source image pixels as the
        blending factor. If the source and destination regions are different sizes, the
        image will be automatically resized to match the destination size. If the
        ``src`` parameter is not used, the display window is used as the source image.

        As of release 0149, this function ignores ``image_mode()``.
"""
        return self._instance.blend(*args)

    @overload
    def copy(self) -> Py5Image:
        """Copies a region of pixels from one image into another.

        Underlying Java method: PImage.copy

        Methods
        -------

        You can use any of the following signatures:

         * copy() -> Py5Image
         * copy(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None
         * copy(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destination's upper left corner

        dy: int
            Y coordinate of the destination's upper left corner

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image.

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Copies a region of pixels from one image into another. If the source and
        destination regions aren't the same size, it will automatically resize source
        pixels to fit the specified target region. No alpha information is used in the
        process, however if the source image has an alpha channel set, it will be copied
        as well.

        As of release 0149, this function ignores ``image_mode()``.
"""
        pass

    @overload
    def copy(self, sx: int, sy: int, sw: int, sh: int,
             dx: int, dy: int, dw: int, dh: int, /) -> None:
        """Copies a region of pixels from one image into another.

        Underlying Java method: PImage.copy

        Methods
        -------

        You can use any of the following signatures:

         * copy() -> Py5Image
         * copy(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None
         * copy(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destination's upper left corner

        dy: int
            Y coordinate of the destination's upper left corner

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image.

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Copies a region of pixels from one image into another. If the source and
        destination regions aren't the same size, it will automatically resize source
        pixels to fit the specified target region. No alpha information is used in the
        process, however if the source image has an alpha channel set, it will be copied
        as well.

        As of release 0149, this function ignores ``image_mode()``.
"""
        pass

    @overload
    def copy(self, src: Py5Image, sx: int, sy: int, sw: int,
             sh: int, dx: int, dy: int, dw: int, dh: int, /) -> None:
        """Copies a region of pixels from one image into another.

        Underlying Java method: PImage.copy

        Methods
        -------

        You can use any of the following signatures:

         * copy() -> Py5Image
         * copy(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None
         * copy(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destination's upper left corner

        dy: int
            Y coordinate of the destination's upper left corner

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image.

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Copies a region of pixels from one image into another. If the source and
        destination regions aren't the same size, it will automatically resize source
        pixels to fit the specified target region. No alpha information is used in the
        process, however if the source image has an alpha channel set, it will be copied
        as well.

        As of release 0149, this function ignores ``image_mode()``.
"""
        pass

    def copy(self, *args):
        """Copies a region of pixels from one image into another.

        Underlying Java method: PImage.copy

        Methods
        -------

        You can use any of the following signatures:

         * copy() -> Py5Image
         * copy(src: Py5Image, sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None
         * copy(sx: int, sy: int, sw: int, sh: int, dx: int, dy: int, dw: int, dh: int) -> None

        Parameters
        ----------

        dh: int
            destination image height

        dw: int
            destination image width

        dx: int
            X coordinate of the destination's upper left corner

        dy: int
            Y coordinate of the destination's upper left corner

        sh: int
            source image height

        src: Py5Image
            an image variable referring to the source image.

        sw: int
            source image width

        sx: int
            X coordinate of the source's upper left corner

        sy: int
            Y coordinate of the source's upper left corner

        Notes
        -----

        Copies a region of pixels from one image into another. If the source and
        destination regions aren't the same size, it will automatically resize source
        pixels to fit the specified target region. No alpha information is used in the
        process, however if the source image has an alpha channel set, it will be copied
        as well.

        As of release 0149, this function ignores ``image_mode()``.
"""
        return self._instance.copy(*args)

    @overload
    def apply_filter(self, kind: int, /) -> None:
        """Filters the image as defined by one of the following modes:  THRESHOLD Converts
        the image to black and white pixels depending if they are above or below the
        threshold defined by the level parameter. The parameter must be between 0.0
        (black) and 1.0 (white). If no level is specified, 0.5 is used.  GRAY Converts
        any colors in the image to grayscale equivalents. No parameter is used.  OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.  INVERT Sets
        each pixel to its inverse value. No parameter is used.  POSTERIZE Limits each
        channel of the image to the number of colors specified as the parameter. The
        parameter can be set to values between 2 and 255, but results are most
        noticeable in the lower ranges.  BLUR Executes a Gaussian blur with the level
        parameter specifying the extent of the blurring. If no parameter is used, the
        blur is equivalent to Gaussian blur of radius 1. Larger values increase the
        blur.  ERODE Reduces the light areas. No parameter is used.  DILATE Increases
        the light areas. No parameter is used.

        Underlying Java method: PImage.filter

        Methods
        -------

        You can use any of the following signatures:

         * apply_filter(kind: int) -> None
         * apply_filter(kind: int, param: float) -> None

        Parameters
        ----------

        kind: int
            Either THRESHOLD, GRAY, OPAQUE, INVERT, POSTERIZE, BLUR, ERODE, or DILATE

        param: float
            unique for each, see above

        Notes
        -----

        Filters the image as defined by one of the following modes:

        THRESHOLD
        Converts the image to black and white pixels depending if they are above or
        below the threshold defined by the level parameter. The parameter must be
        between 0.0 (black) and 1.0 (white). If no level is specified, 0.5 is used.

        GRAY
        Converts any colors in the image to grayscale equivalents. No parameter is used.

        OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.

        INVERT
        Sets each pixel to its inverse value. No parameter is used.

        POSTERIZE
        Limits each channel of the image to the number of colors specified as the
        parameter. The parameter can be set to values between 2 and 255, but results are
        most noticeable in the lower ranges.

        BLUR
        Executes a Gaussian blur with the level parameter specifying the extent of the
        blurring. If no parameter is used, the blur is equivalent to Gaussian blur of
        radius 1. Larger values increase the blur.

        ERODE
        Reduces the light areas. No parameter is used.

        DILATE
        Increases the light areas. No parameter is used.
"""
        pass

    @overload
    def apply_filter(self, kind: int, param: float, /) -> None:
        """Filters the image as defined by one of the following modes:  THRESHOLD Converts
        the image to black and white pixels depending if they are above or below the
        threshold defined by the level parameter. The parameter must be between 0.0
        (black) and 1.0 (white). If no level is specified, 0.5 is used.  GRAY Converts
        any colors in the image to grayscale equivalents. No parameter is used.  OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.  INVERT Sets
        each pixel to its inverse value. No parameter is used.  POSTERIZE Limits each
        channel of the image to the number of colors specified as the parameter. The
        parameter can be set to values between 2 and 255, but results are most
        noticeable in the lower ranges.  BLUR Executes a Gaussian blur with the level
        parameter specifying the extent of the blurring. If no parameter is used, the
        blur is equivalent to Gaussian blur of radius 1. Larger values increase the
        blur.  ERODE Reduces the light areas. No parameter is used.  DILATE Increases
        the light areas. No parameter is used.

        Underlying Java method: PImage.filter

        Methods
        -------

        You can use any of the following signatures:

         * apply_filter(kind: int) -> None
         * apply_filter(kind: int, param: float) -> None

        Parameters
        ----------

        kind: int
            Either THRESHOLD, GRAY, OPAQUE, INVERT, POSTERIZE, BLUR, ERODE, or DILATE

        param: float
            unique for each, see above

        Notes
        -----

        Filters the image as defined by one of the following modes:

        THRESHOLD
        Converts the image to black and white pixels depending if they are above or
        below the threshold defined by the level parameter. The parameter must be
        between 0.0 (black) and 1.0 (white). If no level is specified, 0.5 is used.

        GRAY
        Converts any colors in the image to grayscale equivalents. No parameter is used.

        OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.

        INVERT
        Sets each pixel to its inverse value. No parameter is used.

        POSTERIZE
        Limits each channel of the image to the number of colors specified as the
        parameter. The parameter can be set to values between 2 and 255, but results are
        most noticeable in the lower ranges.

        BLUR
        Executes a Gaussian blur with the level parameter specifying the extent of the
        blurring. If no parameter is used, the blur is equivalent to Gaussian blur of
        radius 1. Larger values increase the blur.

        ERODE
        Reduces the light areas. No parameter is used.

        DILATE
        Increases the light areas. No parameter is used.
"""
        pass

    def apply_filter(self, *args):
        """Filters the image as defined by one of the following modes:  THRESHOLD Converts
        the image to black and white pixels depending if they are above or below the
        threshold defined by the level parameter. The parameter must be between 0.0
        (black) and 1.0 (white). If no level is specified, 0.5 is used.  GRAY Converts
        any colors in the image to grayscale equivalents. No parameter is used.  OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.  INVERT Sets
        each pixel to its inverse value. No parameter is used.  POSTERIZE Limits each
        channel of the image to the number of colors specified as the parameter. The
        parameter can be set to values between 2 and 255, but results are most
        noticeable in the lower ranges.  BLUR Executes a Gaussian blur with the level
        parameter specifying the extent of the blurring. If no parameter is used, the
        blur is equivalent to Gaussian blur of radius 1. Larger values increase the
        blur.  ERODE Reduces the light areas. No parameter is used.  DILATE Increases
        the light areas. No parameter is used.

        Underlying Java method: PImage.filter

        Methods
        -------

        You can use any of the following signatures:

         * apply_filter(kind: int) -> None
         * apply_filter(kind: int, param: float) -> None

        Parameters
        ----------

        kind: int
            Either THRESHOLD, GRAY, OPAQUE, INVERT, POSTERIZE, BLUR, ERODE, or DILATE

        param: float
            unique for each, see above

        Notes
        -----

        Filters the image as defined by one of the following modes:

        THRESHOLD
        Converts the image to black and white pixels depending if they are above or
        below the threshold defined by the level parameter. The parameter must be
        between 0.0 (black) and 1.0 (white). If no level is specified, 0.5 is used.

        GRAY
        Converts any colors in the image to grayscale equivalents. No parameter is used.

        OPAQUE
        Sets the alpha channel to entirely opaque. No parameter is used.

        INVERT
        Sets each pixel to its inverse value. No parameter is used.

        POSTERIZE
        Limits each channel of the image to the number of colors specified as the
        parameter. The parameter can be set to values between 2 and 255, but results are
        most noticeable in the lower ranges.

        BLUR
        Executes a Gaussian blur with the level parameter specifying the extent of the
        blurring. If no parameter is used, the blur is equivalent to Gaussian blur of
        radius 1. Larger values increase the blur.

        ERODE
        Reduces the light areas. No parameter is used.

        DILATE
        Increases the light areas. No parameter is used.
"""
        return self._instance.filter(*args)

    @overload
    def get(self) -> Py5Image:
        """Reads the color of any pixel or grabs a section of an image.

        Underlying Java method: PImage.get

        Methods
        -------

        You can use any of the following signatures:

         * get() -> Py5Image
         * get(x: int, y: int) -> int
         * get(x: int, y: int, w: int, h: int) -> Py5Image

        Parameters
        ----------

        h: int
            height of pixel rectangle to get

        w: int
            width of pixel rectangle to get

        x: int
            x-coordinate of the pixel

        y: int
            y-coordinate of the pixel

        Notes
        -----

        Reads the color of any pixel or grabs a section of an image. If no parameters
        are specified, the entire image is returned. Use the ``x`` and ``y`` parameters
        to get the value of one pixel. Get a section of the display window by specifying
        an additional ``width`` and ``height`` parameter. When getting an image, the
        ``x`` and ``y`` parameters define the coordinates for the upper-left corner of
        the image, regardless of the current ``image_mode()``.

        If the pixel requested is outside of the image window, black is returned. The
        numbers returned are scaled according to the current color ranges, but only RGB
        values are returned by this function. For example, even though you may have
        drawn a shape with ``color_mode(HSB)``, the numbers returned will be in RGB
        format.

        Getting the color of a single pixel with ``get(x, y)`` is easy, but not as fast
        as grabbing the data directly from ``pixels[]``. The equivalent statement to
        ``get(x, y)`` using ``pixels[]`` is ``pixels[y*width+x]``. See the reference for
        ``pixels[]`` for more information.
"""
        pass

    @overload
    def get(self, x: int, y: int, /) -> int:
        """Reads the color of any pixel or grabs a section of an image.

        Underlying Java method: PImage.get

        Methods
        -------

        You can use any of the following signatures:

         * get() -> Py5Image
         * get(x: int, y: int) -> int
         * get(x: int, y: int, w: int, h: int) -> Py5Image

        Parameters
        ----------

        h: int
            height of pixel rectangle to get

        w: int
            width of pixel rectangle to get

        x: int
            x-coordinate of the pixel

        y: int
            y-coordinate of the pixel

        Notes
        -----

        Reads the color of any pixel or grabs a section of an image. If no parameters
        are specified, the entire image is returned. Use the ``x`` and ``y`` parameters
        to get the value of one pixel. Get a section of the display window by specifying
        an additional ``width`` and ``height`` parameter. When getting an image, the
        ``x`` and ``y`` parameters define the coordinates for the upper-left corner of
        the image, regardless of the current ``image_mode()``.

        If the pixel requested is outside of the image window, black is returned. The
        numbers returned are scaled according to the current color ranges, but only RGB
        values are returned by this function. For example, even though you may have
        drawn a shape with ``color_mode(HSB)``, the numbers returned will be in RGB
        format.

        Getting the color of a single pixel with ``get(x, y)`` is easy, but not as fast
        as grabbing the data directly from ``pixels[]``. The equivalent statement to
        ``get(x, y)`` using ``pixels[]`` is ``pixels[y*width+x]``. See the reference for
        ``pixels[]`` for more information.
"""
        pass

    @overload
    def get(self, x: int, y: int, w: int, h: int, /) -> Py5Image:
        """Reads the color of any pixel or grabs a section of an image.

        Underlying Java method: PImage.get

        Methods
        -------

        You can use any of the following signatures:

         * get() -> Py5Image
         * get(x: int, y: int) -> int
         * get(x: int, y: int, w: int, h: int) -> Py5Image

        Parameters
        ----------

        h: int
            height of pixel rectangle to get

        w: int
            width of pixel rectangle to get

        x: int
            x-coordinate of the pixel

        y: int
            y-coordinate of the pixel

        Notes
        -----

        Reads the color of any pixel or grabs a section of an image. If no parameters
        are specified, the entire image is returned. Use the ``x`` and ``y`` parameters
        to get the value of one pixel. Get a section of the display window by specifying
        an additional ``width`` and ``height`` parameter. When getting an image, the
        ``x`` and ``y`` parameters define the coordinates for the upper-left corner of
        the image, regardless of the current ``image_mode()``.

        If the pixel requested is outside of the image window, black is returned. The
        numbers returned are scaled according to the current color ranges, but only RGB
        values are returned by this function. For example, even though you may have
        drawn a shape with ``color_mode(HSB)``, the numbers returned will be in RGB
        format.

        Getting the color of a single pixel with ``get(x, y)`` is easy, but not as fast
        as grabbing the data directly from ``pixels[]``. The equivalent statement to
        ``get(x, y)`` using ``pixels[]`` is ``pixels[y*width+x]``. See the reference for
        ``pixels[]`` for more information.
"""
        pass

    def get(self, *args):
        """Reads the color of any pixel or grabs a section of an image.

        Underlying Java method: PImage.get

        Methods
        -------

        You can use any of the following signatures:

         * get() -> Py5Image
         * get(x: int, y: int) -> int
         * get(x: int, y: int, w: int, h: int) -> Py5Image

        Parameters
        ----------

        h: int
            height of pixel rectangle to get

        w: int
            width of pixel rectangle to get

        x: int
            x-coordinate of the pixel

        y: int
            y-coordinate of the pixel

        Notes
        -----

        Reads the color of any pixel or grabs a section of an image. If no parameters
        are specified, the entire image is returned. Use the ``x`` and ``y`` parameters
        to get the value of one pixel. Get a section of the display window by specifying
        an additional ``width`` and ``height`` parameter. When getting an image, the
        ``x`` and ``y`` parameters define the coordinates for the upper-left corner of
        the image, regardless of the current ``image_mode()``.

        If the pixel requested is outside of the image window, black is returned. The
        numbers returned are scaled according to the current color ranges, but only RGB
        values are returned by this function. For example, even though you may have
        drawn a shape with ``color_mode(HSB)``, the numbers returned will be in RGB
        format.

        Getting the color of a single pixel with ``get(x, y)`` is easy, but not as fast
        as grabbing the data directly from ``pixels[]``. The equivalent statement to
        ``get(x, y)`` using ``pixels[]`` is ``pixels[y*width+x]``. See the reference for
        ``pixels[]`` for more information.
"""
        return self._instance.get(*args)

    def load_pixels(self) -> None:
        """Loads the pixel data for the image into its ``pixels[]`` array.

        Underlying Java method: PImage.loadPixels

        Notes
        -----

        Loads the pixel data for the image into its ``pixels[]`` array. This function
        must always be called before reading from or writing to ``pixels[]``.
"""
        return self._instance.loadPixels()

    @overload
    def mask(self, mask_array: JArray(JInt), /) -> None:
        """Masks part of an image from displaying by loading another image and using it as
        an alpha channel.

        Underlying Java method: PImage.mask

        Methods
        -------

        You can use any of the following signatures:

         * mask(img: Py5Image) -> None
         * mask(mask_array: JArray(JInt)) -> None

        Parameters
        ----------

        img: Py5Image
            image to use as the mask

        mask_array: JArray(JInt)
            array of integers used as the alpha channel, needs to be the same length as the image's pixel array.

        Notes
        -----

        Masks part of an image from displaying by loading another image and using it as
        an alpha channel. This mask image should only contain grayscale data, but only
        the blue color channel is used. The mask image needs to be the same size as the
        image to which it is applied.

        In addition to using a mask image, an integer array containing the alpha channel
        data can be specified directly. This method is useful for creating dynamically
        generated alpha masks. This array must be of the same length as the target
        image's pixels array and should contain only grayscale data of values between
        0-255.
"""
        pass

    @overload
    def mask(self, img: Py5Image, /) -> None:
        """Masks part of an image from displaying by loading another image and using it as
        an alpha channel.

        Underlying Java method: PImage.mask

        Methods
        -------

        You can use any of the following signatures:

         * mask(img: Py5Image) -> None
         * mask(mask_array: JArray(JInt)) -> None

        Parameters
        ----------

        img: Py5Image
            image to use as the mask

        mask_array: JArray(JInt)
            array of integers used as the alpha channel, needs to be the same length as the image's pixel array.

        Notes
        -----

        Masks part of an image from displaying by loading another image and using it as
        an alpha channel. This mask image should only contain grayscale data, but only
        the blue color channel is used. The mask image needs to be the same size as the
        image to which it is applied.

        In addition to using a mask image, an integer array containing the alpha channel
        data can be specified directly. This method is useful for creating dynamically
        generated alpha masks. This array must be of the same length as the target
        image's pixels array and should contain only grayscale data of values between
        0-255.
"""
        pass

    def mask(self, *args):
        """Masks part of an image from displaying by loading another image and using it as
        an alpha channel.

        Underlying Java method: PImage.mask

        Methods
        -------

        You can use any of the following signatures:

         * mask(img: Py5Image) -> None
         * mask(mask_array: JArray(JInt)) -> None

        Parameters
        ----------

        img: Py5Image
            image to use as the mask

        mask_array: JArray(JInt)
            array of integers used as the alpha channel, needs to be the same length as the image's pixel array.

        Notes
        -----

        Masks part of an image from displaying by loading another image and using it as
        an alpha channel. This mask image should only contain grayscale data, but only
        the blue color channel is used. The mask image needs to be the same size as the
        image to which it is applied.

        In addition to using a mask image, an integer array containing the alpha channel
        data can be specified directly. This method is useful for creating dynamically
        generated alpha masks. This array must be of the same length as the target
        image's pixels array and should contain only grayscale data of values between
        0-255.
"""
        return self._instance.mask(*args)

    @overload
    def update_pixels(self) -> None:
        """Updates the image with the data in its ``pixels[]`` array.

        Underlying Java method: PImage.updatePixels

        Methods
        -------

        You can use any of the following signatures:

         * update_pixels() -> None
         * update_pixels(x: int, y: int, w: int, h: int) -> None

        Parameters
        ----------

        h: int
            height

        w: int
            width

        x: int
            x-coordinate of the upper-left corner

        y: int
            y-coordinate of the upper-left corner

        Notes
        -----

        Updates the image with the data in its ``pixels[]`` array. Use in conjunction
        with ``load_pixels()``. If you're only reading pixels from the array, there's no
        need to call ``update_pixels()``.
"""
        pass

    @overload
    def update_pixels(self, x: int, y: int, w: int, h: int, /) -> None:
        """Updates the image with the data in its ``pixels[]`` array.

        Underlying Java method: PImage.updatePixels

        Methods
        -------

        You can use any of the following signatures:

         * update_pixels() -> None
         * update_pixels(x: int, y: int, w: int, h: int) -> None

        Parameters
        ----------

        h: int
            height

        w: int
            width

        x: int
            x-coordinate of the upper-left corner

        y: int
            y-coordinate of the upper-left corner

        Notes
        -----

        Updates the image with the data in its ``pixels[]`` array. Use in conjunction
        with ``load_pixels()``. If you're only reading pixels from the array, there's no
        need to call ``update_pixels()``.
"""
        pass

    def update_pixels(self, *args):
        """Updates the image with the data in its ``pixels[]`` array.

        Underlying Java method: PImage.updatePixels

        Methods
        -------

        You can use any of the following signatures:

         * update_pixels() -> None
         * update_pixels(x: int, y: int, w: int, h: int) -> None

        Parameters
        ----------

        h: int
            height

        w: int
            width

        x: int
            x-coordinate of the upper-left corner

        y: int
            y-coordinate of the upper-left corner

        Notes
        -----

        Updates the image with the data in its ``pixels[]`` array. Use in conjunction
        with ``load_pixels()``. If you're only reading pixels from the array, there's no
        need to call ``update_pixels()``.
"""
        return self._instance.updatePixels(*args)



from typing import overload, Union

import numpy as np
from numpy.random import MT19937
from numpy.random import RandomState, SeedSequence

import noise


class MathMixin:

    _rs = RandomState(MT19937(SeedSequence()))

    SIMPLEX_NOISE = 1  # CODEBUILDER INCLUDE
    PERLIN_NOISE = 2  # CODEBUILDER INCLUDE
    _NOISE_MODE = SIMPLEX_NOISE
    _NOISE_SEED = 0
    _NOISE_OCTAVES = 4
    _NOISE_PERSISTENCE = 0.5
    _NOISE_LACUNARITY = 2.0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # *** BEGIN METHODS ***

    @classmethod
    def sin(cls, angle: float) -> float:
        """new template no description.

        Parameters
        ----------

        angle: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.sin(angle)

    @classmethod
    def cos(cls, angle: float) -> float:
        """new template no description.

        Parameters
        ----------

        angle: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.cos(angle)

    @classmethod
    def tan(cls, angle: float) -> float:
        """new template no description.

        Parameters
        ----------

        angle: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.tan(angle)

    @classmethod
    def asin(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.arcsin(value)

    @classmethod
    def acos(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.arccos(value)

    @classmethod
    def atan(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.arctan(value)

    @classmethod
    def atan2(cls, y: float, x: float) -> float:
        """new template no description.

        Parameters
        ----------

        x: float
            missing variable description

        y: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.arctan2(y, x)

    @classmethod
    def degrees(cls, radians: float) -> float:
        """new template no description.

        Parameters
        ----------

        radians: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.degrees(radians)

    @classmethod
    def radians(cls, degrees: float) -> float:
        """new template no description.

        Parameters
        ----------

        degrees: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.radians(degrees)

    @classmethod
    def constrain(cls, amt: float, low: float, high: float) -> float:
        """new template no description.

        Parameters
        ----------

        amt: float
            missing variable description

        high: float
            missing variable description

        low: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.where(amt < low, low, np.where(amt > high, high, amt))

    @classmethod
    def remap(
            cls,
            value: float,
            start1: float,
            stop1: float,
            start2: float,
            stop2: float) -> float:
        """new template no description.

        Parameters
        ----------

        start1: float
            missing variable description

        start2: float
            missing variable description

        stop1: float
            missing variable description

        stop2: float
            missing variable description

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return start2 + (stop2 - start2) * \
            ((value - start1) / (stop1 - start1))

    @classmethod
    def dist(cls, *args: float) -> float:
        """new template no description.

        Parameters
        ----------

        args: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        p1 = args[:(len(args) // 2)]
        p2 = args[(len(args) // 2):]
        assert len(p1) == len(p2)
        return sum([(a - b)**2 for a, b in zip(p1, p2)])**0.5

    @classmethod
    def lerp(cls, start: float, stop: float, amt: float) -> float:
        """new template no description.

        Parameters
        ----------

        amt: float
            missing variable description

        start: float
            missing variable description

        stop: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return amt * (stop - start) + start

    @classmethod
    def mag(cls, *args: float) -> float:
        """new template no description.

        Parameters
        ----------

        args: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return sum([x * x for x in args])**0.5

    @classmethod
    def norm(cls, value: float, start: float, stop: float) -> float:
        """new template no description.

        Parameters
        ----------

        start: float
            missing variable description

        stop: float
            missing variable description

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return (value - start) / (stop - start)

    @classmethod
    def sq(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return value * value

    @classmethod
    def sqrt(cls, value: float) -> Union[float, complex]:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return value**0.5

    @classmethod
    def floor(cls, value: float) -> int:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return int(np.floor(value))

    @classmethod
    def ceil(cls, value: float) -> int:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return int(np.ceil(value))

    @classmethod
    def exp(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.exp(value)

    @classmethod
    def log(cls, value: float) -> float:
        """new template no description.

        Parameters
        ----------

        value: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return np.log(value)

    @overload
    def random(cls, high: float) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * random(high: float) -> float
         * random(low: float, high: float) -> float

        Parameters
        ----------

        high: float
            upper limit

        low: float
            lower limit

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def random(cls, low: float, high: float) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * random(high: float) -> float
         * random(low: float, high: float) -> float

        Parameters
        ----------

        high: float
            upper limit

        low: float
            lower limit

        Notes
        -----

        new template no description.
"""
        pass

    @classmethod
    def random_seed(cls, seed: int) -> None:
        """new template no description.

        Parameters
        ----------

        seed: int
            seed value

        Notes
        -----

        new template no description.
"""
        cls._rs = RandomState(MT19937(SeedSequence(seed)))

    @classmethod
    def random(cls, *args: float) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * random(high: float) -> float
         * random(low: float, high: float) -> float

        Parameters
        ----------

        high: float
            upper limit

        low: float
            lower limit

        Notes
        -----

        new template no description.
"""
        if len(args) == 1:
            high = args[0]
            if isinstance(high, (int, float)):
                return high * cls._rs.rand()
        elif len(args) == 2:
            low, high = args
            if isinstance(
                    low, (int, float)) and isinstance(
                    high, (int, float)):
                return low + (high - low) * cls._rs.rand()

        types = ','.join([type(a).__name__ for a in args])
        raise TypeError(
            f'No matching overloads found for Sketch.random({types})')

    @classmethod
    def random_gaussian(cls) -> float:
        """new template no description.

        Notes
        -----

        new template no description.
"""
        return cls._rs.randn()

    @overload
    def noise(cls, x, **kwargs) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * noise(x, kwargs) -> float
         * noise(x, y, kwargs) -> float
         * noise(x, y, z, kwargs) -> float
         * noise(x, y, z, w, kwargs) -> float

        Parameters
        ----------

        kwargs
            missing variable description

        w
            missing variable description

        x
            x-coordinate in noise space

        y
            y-coordinate in noise space

        z
            z-coordinate in noise space

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def noise(cls, x, y, **kwargs) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * noise(x, kwargs) -> float
         * noise(x, y, kwargs) -> float
         * noise(x, y, z, kwargs) -> float
         * noise(x, y, z, w, kwargs) -> float

        Parameters
        ----------

        kwargs
            missing variable description

        w
            missing variable description

        x
            x-coordinate in noise space

        y
            y-coordinate in noise space

        z
            z-coordinate in noise space

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def noise(cls, x, y, z, **kwargs) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * noise(x, kwargs) -> float
         * noise(x, y, kwargs) -> float
         * noise(x, y, z, kwargs) -> float
         * noise(x, y, z, w, kwargs) -> float

        Parameters
        ----------

        kwargs
            missing variable description

        w
            missing variable description

        x
            x-coordinate in noise space

        y
            y-coordinate in noise space

        z
            z-coordinate in noise space

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def noise(cls, x, y, z, w, **kwargs) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * noise(x, kwargs) -> float
         * noise(x, y, kwargs) -> float
         * noise(x, y, z, kwargs) -> float
         * noise(x, y, z, w, kwargs) -> float

        Parameters
        ----------

        kwargs
            missing variable description

        w
            missing variable description

        x
            x-coordinate in noise space

        y
            y-coordinate in noise space

        z
            z-coordinate in noise space

        Notes
        -----

        new template no description.
"""
        pass

    @classmethod
    def noise(cls, *args, **kwargs) -> float:
        """new template no description.

        Methods
        -------

        You can use any of the following signatures:

         * noise(x, kwargs) -> float
         * noise(x, y, kwargs) -> float
         * noise(x, y, z, kwargs) -> float
         * noise(x, y, z, w, kwargs) -> float

        Parameters
        ----------

        kwargs
            missing variable description

        w
            missing variable description

        x
            x-coordinate in noise space

        y
            y-coordinate in noise space

        z
            z-coordinate in noise space

        Notes
        -----

        new template no description.
"""
        len_args = len(args)
        noise_args = {
            'octaves': cls._NOISE_OCTAVES,
            'persistence': cls._NOISE_PERSISTENCE,
            'lacunarity': cls._NOISE_LACUNARITY,
            'base': cls._NOISE_SEED,
            # this will override other parameters if specified by the user
            **kwargs
        }
        noisef = lambda *x, **_: x[0]
        if cls._NOISE_MODE == cls.PERLIN_NOISE:
            if len_args not in [1, 2, 3]:
                raise RuntimeError(
                    'Sorry, this noise function can only generate 1D, 2D, or 3D Perlin noise.')
            noisef = {
                1: noise.pnoise1,
                2: noise.pnoise2,
                3: noise.pnoise3}[len_args]
        elif cls._NOISE_MODE == cls.SIMPLEX_NOISE:
            if len_args not in [1, 2, 3, 4]:
                raise RuntimeError(
                    'Sorry, this noise function can only generate 1D, 2D, 3D, or 4D Simplex noise.')
            noisef = {
                1: noise.snoise2,
                2: noise.snoise2,
                3: noise.snoise3,
                4: noise.snoise4}[len_args]
            if len_args == 1:
                args = args[0], 0
            if len_args in [3, 4]:
                del noise_args['base']
        if any(isinstance(v, np.ndarray) for v in args):
            noisef = np.vectorize(noisef)
        return noisef(*args, **noise_args)

    @classmethod
    def noise_mode(cls, mode: int) -> None:
        """new template no description.

        Parameters
        ----------

        mode: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        if mode in [cls.PERLIN_NOISE, cls.SIMPLEX_NOISE]:
            cls._NOISE_MODE = mode

    @classmethod
    def noise_detail(cls, octaves: float = None, persistence: float = None,
                     lacunarity: float = None) -> None:
        """new template no description.

        Parameters
        ----------

        lacunarity: float
            missing variable description

        octaves: float
            missing variable description

        persistence: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        if octaves:
            cls._NOISE_OCTAVES = octaves
        if persistence:
            cls._NOISE_PERSISTENCE = persistence
        if lacunarity:
            cls._NOISE_LACUNARITY = lacunarity

    @classmethod
    def noise_seed(cls, seed: float) -> None:
        """new template no description.

        Parameters
        ----------

        seed: float
            seed value

        Notes
        -----

        new template no description.
"""
        cls._NOISE_SEED = seed



import cmd
import argparse
import platform
import glob
from pathlib import Path

import py5_tools


parser = argparse.ArgumentParser(description="py5 command tool",
                                 epilog="this is the epilog")


LIBRARY_TEMPLATE = """[{id}] Name: {name}
Author: {authors}
{sentence}
{categories}
{paragraph}"""


class Py5Cmd(cmd.Cmd):

    def __init__(self):
        super().__init__()
        self._libraries = py5_tools.ProcessingLibraryInfo()
        self._running_sketches = []

    prompt = 'py5: '
    intro = "Welcome to the py5 command tool."

    def _print_library_info(self, info):
        info = info.T.to_dict()[info.index[0]]

        return LIBRARY_TEMPLATE.format(**info)

    def do_list_categories(self, line):
        for c in self._libraries.categories:
            print(c)

    def do_show_category(self, line):
        category_libraries = self._libraries.get_library_info(
            category=line).sort_values('id')

        for _, info in category_libraries.iterrows():
            print(LIBRARY_TEMPLATE.format(**info).strip() + '\n')

    def complete_show_category(self, text, line, begidx, endidx):
        if not text:
            completions = self._libraries.categories
        else:
            completions = [
                c for c in self._libraries.categories if c.startswith(text)]

        return completions

    def do_run_sketch(self, line):
        if line:
            try:
                new_process = platform.system() != 'Windows'
                p = py5_tools.run.run_sketch(line, new_process=new_process)
                self._running_sketches.append(p)
            except Exception as e:
                print(e)

    def complete_run_sketch(self, text, line, begidx, endidx):
        path = line[10:].strip()
        completions = []
        for p in glob.glob(path + '*'):
            completions.append(p[(len(path) - len(text)):] +
                               ('/' if Path(p).is_dir() else ''))
        return completions

    def do_get_library(self, line):
        try:
            self._libraries.download_zip('jars', library_name=line)
        except Exception as e:
            print(e)

    def complete_get_library(self, text, line, begidx, endidx):
        if not text:
            completions = self._libraries.names
        else:
            completions = [
                n for n in self._libraries.names if n.startswith(text)]

        return completions

    def emptyline(self):
        return None

    def do_exit(self, line):
        return True

    def do_EOF(self, line):
        for p in self._running_sketches:
            p.terminate()
            p.join()

        return True


def main():
    # args = parser.parse_args()
    py5cmd = Py5Cmd()
    py5cmd.cmdloop()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***
import functools
from typing import overload, List  # noqa

import jpype
from jpype import JException, JArray, JString  # noqa

from .shape import Py5Shape, _return_py5shape  # noqa


def _return_py5font(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        return Py5Font(f(self_, *args))
    return decorated


def _load_py5font(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        # TODO: for load_font this prints a Java exception to strerr if the
        # file cannot be found or read
        try:
            ret = f(self_, *args)
        except JException as e:
            msg = e.message()
        else:
            if ret is None:
                msg = 'font file is missing or inaccessible.'
            else:
                return Py5Font(ret)
        raise RuntimeError('cannot load font file ' +
                           str(args[0]) + '. error message: ' + msg)
    return decorated


def _return_list_str(f):
    @functools.wraps(f)
    def decorated(cls_, *args):
        return [str(x) for x in f(cls_, *args) or []]
    return decorated


class Py5Font:
    """PFont is the font class for Processing.

    Underlying Java class: PFont.PFont

    Notes
    -----

    PFont is the font class for Processing. To create a font to use with Processing,
    select "Create Font..." from the Tools menu. This will create a font in the
    format Processing requires and also adds it to the current sketch's data
    directory. Processing displays fonts using the .vlw font format, which uses
    images for each letter, rather than defining them through vector data. The
    ``load_font()`` function constructs a new font and ``text_font()`` makes a font
    active. The ``list()`` method creates a list of the fonts installed on the
    computer, which is useful information to use with the ``create_font()`` function
    for dynamically converting fonts into a format to use with Processing.

    To create a new font dynamically, use the ``create_font()`` function. Do not use
    the syntax ``new Py5Font()``.
"""

    _cls = jpype.JClass('processing.core.PFont')
    CHARSET = _cls.CHARSET

    def __init__(self, pfont):
        self._instance = pfont

    def ascent(self) -> float:
        """new template no description.

        Underlying Java method: PFont.ascent

        Notes
        -----

        new template no description.
"""
        return self._instance.ascent()

    def descent(self) -> float:
        """new template no description.

        Underlying Java method: PFont.descent

        Notes
        -----

        new template no description.
"""
        return self._instance.descent()

    def get_default_size(self) -> int:
        """new template no description.

        Underlying Java method: PFont.getDefaultSize

        Notes
        -----

        new template no description.
"""
        return self._instance.getDefaultSize()

    def get_glyph_count(self) -> int:
        """new template no description.

        Underlying Java method: PFont.getGlyphCount

        Notes
        -----

        new template no description.
"""
        return self._instance.getGlyphCount()

    def get_name(self) -> str:
        """new template no description.

        Underlying Java method: PFont.getName

        Notes
        -----

        new template no description.
"""
        return self._instance.getName()

    def get_post_script_name(self) -> str:
        """new template no description.

        Underlying Java method: PFont.getPostScriptName

        Notes
        -----

        new template no description.
"""
        return self._instance.getPostScriptName()

    @overload
    def get_shape(self, ch: chr, /) -> Py5Shape:
        """new template no description.

        Underlying Java method: PFont.getShape

        Methods
        -------

        You can use any of the following signatures:

         * get_shape(ch: chr) -> Py5Shape
         * get_shape(ch: chr, detail: float) -> Py5Shape

        Parameters
        ----------

        ch: chr
            missing variable description

        detail: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @overload
    def get_shape(self, ch: chr, detail: float, /) -> Py5Shape:
        """new template no description.

        Underlying Java method: PFont.getShape

        Methods
        -------

        You can use any of the following signatures:

         * get_shape(ch: chr) -> Py5Shape
         * get_shape(ch: chr, detail: float) -> Py5Shape

        Parameters
        ----------

        ch: chr
            missing variable description

        detail: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        pass

    @_return_py5shape
    def get_shape(self, *args):
        """new template no description.

        Underlying Java method: PFont.getShape

        Methods
        -------

        You can use any of the following signatures:

         * get_shape(ch: chr) -> Py5Shape
         * get_shape(ch: chr, detail: float) -> Py5Shape

        Parameters
        ----------

        ch: chr
            missing variable description

        detail: float
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.getShape(*args)

    def get_size(self) -> int:
        """new template no description.

        Underlying Java method: PFont.getSize

        Notes
        -----

        new template no description.
"""
        return self._instance.getSize()

    def is_smooth(self) -> bool:
        """new template no description.

        Underlying Java method: PFont.isSmooth

        Notes
        -----

        new template no description.
"""
        return self._instance.isSmooth()

    def is_stream(self) -> bool:
        """new template no description.

        Underlying Java method: PFont.isStream

        Notes
        -----

        new template no description.
"""
        return self._instance.isStream()

    @classmethod
    @_return_list_str
    def list(cls) -> List[str]:
        """Gets a list of the fonts installed on the system.

        Underlying Java method: PFont.list

        Notes
        -----

        Gets a list of the fonts installed on the system. The data is returned as a
        String array. This list provides the names of each font for input into
        ``create_font()``, which allows Processing to dynamically format fonts.
"""
        return cls._cls.list()

    def set_subsetting(self) -> None:
        """new template no description.

        Underlying Java method: PFont.setSubsetting

        Notes
        -----

        new template no description.
"""
        return self._instance.setSubsetting()

    def width(self, c: chr, /) -> float:
        """new template no description.

        Underlying Java method: PFont.width

        Parameters
        ----------

        c: chr
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.width(c)



import argparse

from py5_tools import run


parser = argparse.ArgumentParser(description="Execute py5 sketch",
                                 epilog="this is the epilog")
parser.add_argument(
    action='store',
    dest='sketch_path',
    help='path to py5 sketch')
parser.add_argument('-c', '--classpath', action='store', dest='classpath',
                    help='extra directories to add to classpath')


def main():
    args = parser.parse_args()
    run.run_sketch(args.sketch_path, classpath=args.classpath)


if __name__ == '__main__':
    main()



from .sketch import Sketch


class CreateFontTool(Sketch):

    def __init__(
            self,
            font_name,
            font_size,
            filename=None,
            characters=None,
            pause=True):
        super().__init__()
        self.font_name = font_name
        self.font_size = font_size
        self.pause = pause
        self.filename = filename or f'{font_name}-{font_size}.vlw'
        self.characters = characters

    def settings(self):
        self.size(400, 100, self.P2D)

    def setup(self):
        font = self.create_font(self.font_name, self.font_size)
        self.text_align(self.CENTER)

        self.background(0)
        self.fill(0)

        self.text_font(font)
        characters = self.characters or ''.join(font.CHARSET)
        self.text(characters, self.width / 2, self.height / 2)

        os = self.get_py5applet().createOutput(str(self.filename))
        font._instance.save(os)
        os.close()

        self.fill(255)
        msg = str(font.get_glyph_count()) + \
            ' glyphs written to ' + self.filename
        self.translate(self.width / 2, self.height / 2)
        self.scale(0.95 * self.width / self.text_width(msg))
        self.text(msg, 0, 0)

    def draw(self):
        if self.pause:
            self.no_loop()
        else:
            self.exit_sketch()


def create_font_file(
        font_name: str,
        font_size: int,
        filename: str = None,
        characters: str = None,
        pause: bool = True):
    """missing docstring"""
    vlw_creator = CreateFontTool(font_name, font_size,
                                 filename=filename, characters=characters,
                                 pause=pause)
    vlw_creator.run_sketch(block=pause)



# -*- coding: utf-8 -*-
# *** FORMAT PARAMS ***
import functools
from typing import overload, Any  # noqa

from .image import Py5Image  # noqa


def _return_py5surface(f):
    @functools.wraps(f)
    def decorated(self_, *args):
        return Py5Surface(
            f(self_, *args), getattr(self_, '_pimage_cache', None))
    return decorated


class Py5Surface:
    """new template no description.

    Underlying Java class: PSurface.PSurface

    Notes
    -----

    new template no description.
"""

    def __init__(self, psurface, pimage_cache):
        self._instance = psurface
        self._pimage_cache = pimage_cache

    def get_native(self) -> Any:
        """new template no description.

        Underlying Java method: PSurface.getNative

        Notes
        -----

        new template no description.
"""
        return self._instance.getNative()

    def is_stopped(self) -> bool:
        """new template no description.

        Underlying Java method: PSurface.isStopped

        Notes
        -----

        new template no description.
"""
        return self._instance.isStopped()

    def open_link(self, url: str, /) -> bool:
        """new template no description.

        Underlying Java method: PSurface.openLink

        Parameters
        ----------

        url: str
            the link to open

        Notes
        -----

        new template no description.
"""
        return self._instance.openLink(url)

    def pause_thread(self) -> None:
        """new template no description.

        Underlying Java method: PSurface.pauseThread

        Notes
        -----

        new template no description.
"""
        return self._instance.pauseThread()

    def resume_thread(self) -> None:
        """new template no description.

        Underlying Java method: PSurface.resumeThread

        Notes
        -----

        new template no description.
"""
        return self._instance.resumeThread()

    def set_always_on_top(self, always: bool, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setAlwaysOnTop

        Parameters
        ----------

        always: bool
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setAlwaysOnTop(always)

    def set_icon(self, icon: Py5Image, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setIcon

        Parameters
        ----------

        icon: Py5Image
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setIcon(icon)

    def set_location(self, x: int, y: int, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setLocation

        Parameters
        ----------

        x: int
            missing variable description

        y: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setLocation(x, y)

    def set_resizable(self, resizable: bool, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setResizable

        Parameters
        ----------

        resizable: bool
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setResizable(resizable)

    def set_size(self, width: int, height: int, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setSize

        Parameters
        ----------

        height: int
            missing variable description

        width: int
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setSize(width, height)

    def set_title(self, title: str, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setTitle

        Parameters
        ----------

        title: str
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setTitle(title)

    def set_visible(self, visible: bool, /) -> None:
        """new template no description.

        Underlying Java method: PSurface.setVisible

        Parameters
        ----------

        visible: bool
            missing variable description

        Notes
        -----

        new template no description.
"""
        return self._instance.setVisible(visible)

    def stop_thread(self) -> bool:
        """new template no description.

        Underlying Java method: PSurface.stopThread

        Notes
        -----

        new template no description.
"""
        return self._instance.stopThread()



import re
import io
import zipfile
import requests
from functools import reduce
from pathlib import Path

import pandas as pd


PROCESSING_LIBRARY_URL = 'http://download.processing.org/contribs'

PARAGRAPH_REGEX = re.compile(
    '^paragraph=(.*?)^[a-z]*?=',
    re.DOTALL | re.MULTILINE)


class ProcessingLibraryInfo:

    def __init__(self):
        self._load_data()

    def _load_data(self):
        response = requests.get(PROCESSING_LIBRARY_URL)
        if response.status_code != 200:
            raise RuntimeError(
                f'could not download data file at {PROCESSING_LIBRARY_URL}')

        blocks = [b for b in response.text.split(
            '\n\n') if b.startswith('library')]
        block_lines = [dict([line.split('=', 1)
                             for line in block.split('\n')
                             if line != "library"])
                       for block in blocks]
        df = pd.DataFrame.from_dict(block_lines, dtype="string")
        # lastUpdated is supposed to be the last column
        df = df.iloc[:, :(df.columns.get_loc('lastUpdated') + 1)]
        df.astype({'id': int, 'minRevision': int, 'maxRevision': int})
        df['id'] = df['id'].astype(int)
        df['minRevision'] = df['minRevision'].astype(int)
        df['maxRevision'] = df['maxRevision'].astype(int)
        df['categories'] = df['categories'].apply(lambda x: x.split(','))
        # get paragraph values from raw data because they could be on more than one
        # line or the paragraph could be missing
        df['paragraph'] = [PARAGRAPH_REGEX.findall(b) for b in blocks]
        df['paragraph'] = df['paragraph'].apply(
            lambda x: x[0] if x else '').astype('string')

        self._data = df
        self.categories = sorted(
            reduce(
                lambda x,
                y: x | set(y),
                df['categories'],
                set()))
        self.names = sorted(df['name'])

    def get_library_info(
            self,
            category=None,
            library_name=None,
            library_id=None):
        info = self._data
        # TODO: also make sure minRevision < version < maxRevision
        if category:
            info = info[info['categories'].apply(lambda x: category in x)]
        if library_name:
            info = info[info['name'] == library_name]
        if library_id:
            info = info[info['id'] == library_id]

        return info

    def download_zip(self, dest, library_name=None, library_id=None):
        info = self.get_library_info(
            library_name=library_name,
            library_id=library_id)

        if len(info) == 0:
            raise RuntimeError(f'library not found')
        if len(info) > 1:
            raise RuntimeError(f'more than one library')

        info = info.T.to_dict()[info.index[0]]
        download_url = info['download']

        response = requests.get(download_url)
        if response.status_code != 200:
            raise RuntimeError(f'could not download library at {download_url}')

        with zipfile.ZipFile(io.BytesIO(response.content)) as zf:
            jars = []
            for name in zf.namelist():
                path = Path(name)
                if len(
                        path.parts) > 2 and path.parts[1] == 'library' and path.suffix == '.jar':
                    jars.append(name)
            zf.extractall(dest, jars)


__all__ = ['ProcessingLibraryInfo']
